# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 14 19:11:28 2018

EN 1993-1-8 Rules for connections

@author: kmela
"""

from copy import deepcopy
import math
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as lines
import numpy as np

from metku.eurocodes.en1993.constants import gammaM5, E
from metku.structures.steel.steel_member import SteelMember

def eccentricity(h0,h1,h2,t1,t2,g):
    """ Eccentricity of a K gap joint """
    rt1 = math.radians(t1)
    rt2 = math.radians(t2)
    sin1 = math.sin(rt1)
    sin2 = math.sin(rt2)
    sin12 = math.sin(rt1+rt2)
    return (0.5*h1/sin1 + 0.5*h2/sin2 + g)*sin1*sin2/sin12 - 0.5*h0

def rotate(v,theta=0.0):
    """ Rotates vector 'v' by the angle 'theta'
        Works only in 2D
    """
    
    if len(v) != 2:
        print("Error: length of vector to be rotated must be 2.")
    else:
        trad = np.radians(theta)
        st = np.sin(trad)
        ct = np.cos(trad)
        R = np.array([[ct,-st],[st,ct]])
        
        v = np.asarray(v)
        vrot = R.dot(v)
    
    return vrot

class Line:
    """ Small class for lines in 2D """
    
    def __init__(self,**kwargs):
        """ Constructor
            Possible initialization calls:
                p1, p2: two points through which the line passes
                p1, v: point and direction vector
                
            v ..unit vector
                 
        """
        
        self.p1 = None
        self.p2 = None
        self.v = None
        
        for key, value in kwargs.items():
            if key == 'p1':
                self.p1 = np.asarray(value)
            elif key == 'p2':
                self.p2 = np.asarray(value)
            elif key == 'v':
                self.v = np.asarray(value)
    
        if self.p2 is None:
            self.p2 = self.p1 + self.v
        elif self.v is None:
            self.v = (self.p2-self.p1)/np.linalg.norm(self.p2-self.p1)
            
    def __call__(self,t):
        """ Calculates the coordinates for local coordinate t """
        
        return self.p1 + t*self.v
    
    def __repr__(self):
        
        return f"Line: p1 = ({self.p1[0]:.3f},{self.p1[1]:.3f}), v = ({self.v[0]:.3f},{self.v[1]:.3f})"
            
    def intersect(self,line):
        """ Finds intersection point between two lines
        """
        v = self.v
        p = self.p1
        u = line.v
        r = line.p1
        
        A = np.array([[v[0],-u[0]],[v[1],-u[1]]])
        b = r-p
                        
        t = np.linalg.solve(A,b)
        
        x_intersect = p + t[0]*v
        
        return t, x_intersect
    
    def distance(self,xR):
        """ Finds distance from a point with coordinates in xR
            Returns also the nearest point on the line
        """
        dx = self.p1-xR
        t_close = -self.v.dot(dx)
                
        
        x_closest = self.__call__(t_close)

        distance = np.linalg.norm(x_closest-xR)
        
        return distance, x_closest
    
class RHSJoint:
    """ Class for joints between RHS members """
    
    def __init__(self,chord, N0=0.0, V0=0.0, M0=0.0):
        """ Constructor
            Input:
                
            :param: chord.. can be SteelMember or RHS or SHS class member, for setting the chord profile
            :param: N0 .. magnitude of the largest axial force in the chord around the joint.
            :param: V0 .. magnitude of the largest shear force in the chord around the joint.
            :param: M0 .. magnitude of the largest bending moment in the chord.
        
        
        """
        if not isinstance(chord,SteelMember):
            chord = SteelMember(chord,length=1000)
        
        self.chord = chord
        
        self.N0 = N0
        self.V0 = V0
        self.M0 = M0
        
        
        
        self.r = 0

        self.chord_face_plate = {'lp': 0, 'bp': 0,'tp': 0}
        self.chord_web_plate = {'lp': 0, 'bp': 0,'tp': 0}
             
    
    @property    
    def h0(self):
        return self.chord.profile.H
        
    @property
    def b0(self):
        return self.chord.profile.B
        
    @property
    def t0(self):
        return self.chord.profile.T
    
    @property
    def r0(self):
        return self.chord.profile.R
    
    @property
    def fy0(self):
        return self.chord.fy()

    def gamma(self):
        return 0.5*self.b0/self.t0
        
    def b_straight(self):
        """ Length of the straight part of the chord profile """
        return self.b0-2*self.r0
        
    def beta(self):
        """ Abstract method for beta that depends on the joint type """
        return None
    
    def eval_KN(self):
        """ Evaluate stress function """
        # by default, assume the chord face is in tension: kn = 1.0
        kn = 1.0
        n = self.eval_N()
        b = self.beta()
        if n > 0.0: # compression
            kn = min(1.3-0.4*n/b, 1.0)
        
        return kn
        
    def strength_reduction(self):
        """ Reduce the strength of joint for steel greater than S355
        
            NOTE! Strength reduction factor can be separately specified
            in attribute 'r'.
        """
        
        if self.r == 0:
            r = 1.0
            if self.fy0 <= 355:
                r = 1.0
            elif self.fy0 <= 460:            
                r = 0.9
            else:
                r = 0.8
        else:
            r = self.r
        return r
        
    def eval_N(self):
        """ Evaluate chord stress function """
        # s0Ed > 0 for compression
        s0Ed = self.chord_stress()
        n = s0Ed/self.fy0/gammaM5
        return n
    
    
    def chord_stress(self):
        """ Chord stress ratio """
        A0 = self.chord.profile.A
        Wel0 = self.chord.profile.Wel[0];                 
        # s0Ed is positive, if the stress in the chord
        # is compressive.
        # CAREFUL WITH UNITS HERE!
        s0Ed = -self.N0/A0+self.M0/Wel0
        
        return s0Ed

        
    def new_evalN(self):
        """ Evaluate stress ratio as proposed """
        # s0Ed > 0 for compression
        s0Ed = self.new_chord_axial_stress()
        # n < 0 for compression
        n = -s0Ed/self.fy0/gammaM5
        
        return n
        
    def new_chord_axial_stress(self):
        """ Axial stress in the chord face at the joint
            Based on the new proposal, where plastic bending resistance is 
            used instead of elastic
        """
        A0 = self.chord.profile.A
        Wpl0 = self.chord.profile.Wpl[1]
        # s0Ed < 0 for compression
        s0Ed = -self.N0/A0+self.M0/Wpl0
        return s0Ed

    def info(self):
        """ Prints information on the joint """
        
        print("  Chord: {0:s}".format(self.chord.profile.__repr__()))
        print("    N_0,Ed = {0:4.2f} kN".format(self.N0*1e-3))
        print("    M_0,Ed = {0:4.2f} kNm".format(self.M0*1e-6))
    
    def draw(self,fig=None,ax=None,chord_angle = 0, dx=500):
        """ Draw joint 
            For the main class, only the chord is plotted
            
            input:
                chord_angle .. angle of the chord and the horizontal axis (deg)
                dx .. length (mm) of the chord in each side of the origin
                
            origin is in the middle of the center line of the chord
        """
        
        grey = '0.8'
        if fig is None:
            fig, ax = plt.subplots(1)
        
        # y coordinate of the top face
        y = 0.5*self.h0
        yinner = y-self.t0
        
        ax.hlines([0],-dx,dx,'k','dashdot',zorder=2,linewidth=1)
        ax.hlines([-y,y],-dx,dx,'k','solid',zorder=2,linewidth=1.75)
        ax.hlines([-yinner,yinner],-dx,dx,grey,'solid',zorder=2,linewidth=1.25)
        
        
        ax.set_aspect('equal', 'datalim')
    
class RHSKGapJoint(RHSJoint):
    """ For K and N Gap joints between RHS members """
    
    def __init__(self,chord,braces,angles,gap=20, **kwargs):
        """ Constructor
            input:
                chord .. chord member or profile (Steel Member, RHS or SHS class object)
                braces .. list of brace profiles (RHS or SHS class objects)
                angles .. list of angles (degrees) between the braces and the chord
                gap .. gap between the braces
        """
        RHSJoint.__init__(self,chord, **kwargs)        
        
        if not isinstance(braces[0],SteelMember):
            brace_members = [SteelMember(brace,length=1000) for brace in braces]
            for brace, bsec in zip(brace_members,braces):
                brace.add_section(ned=bsec.Ned)
            braces = brace_members
        
        self.braces = braces
        self.gap = gap
        self.angles = angles
    
    @property
    def NEd(self):
        """ Numpy array of axial forces in braces """
        return np.array([self.braces[0].NEd,self.braces[1].NEd])
    
    @property
    def N1(self):
        """ Axial force in the compression member """
        if self.braces[0].NcEd < 0.0:
            return abs(self.braces[0].NcEd)
        else:
            return self.braces[1].NEd
        
    @property
    def angle1(self):
        """ Angle of the compression member """
        if self.braces[0].NcEd < 0.0:
            return self.angles[0]
        else:
            return self.angles[1]
    
    @property
    def N2(self):
        """ Axial force in the tension member """
        if self.braces[0].NcEd <= 0.0:
            return abs(self.braces[1].NcEd)
        else:
            return self.braces[0].NEd
    
    @property
    def angle2(self):
        """ Angle of the tension member """
        if self.braces[0].Ned <= 0.0:
            return self.angles[1]
        else:
            return self.angles[0]
        
    @property
    def N0gap(self):
        """ Axial force in the gap 
            NOTE: The sign convention is that N0gap < 0 when there is compression
            in the gap and N0gap > 0 for tension.
            
            The expressions are obtained from horizontal equilibrium on the
            N0 side of the joint. When the chord is in compression, 
            N0 is on the same side as the tension brace. When the chord is in tension,
            N0 is on the same side as the compression brace.
        """
        if self.N0 < 0.0:
            # Chord in compression
            N0gap = self.N0+self.N2*np.cos(np.radians(self.angle2))
        else: 
            # Chord in tension            
            N0gap = self.N0+self.N1*np.cos(np.radians(self.angle1))
    
        return N0gap
    
    @property
    def V0gap(self):
        """ Shear force in the gap 
        
            The shear force is in the gap is obtained as the vertical component
            of either of the braces (see vertical equilibrium of the joint).
        """
        
        return abs(self.N1*np.sin(np.radians(self.angle1)))
    
    @property
    def h(self):
        """ Brace heights """
        return np.asarray([self.braces[0].profile.H,self.braces[1].profile.H])
    
    @property
    def b(self):
        """ Brace widths """
        return np.asarray([self.braces[0].profile.B,self.braces[1].profile.B])
    
    @property
    def t(self):
        """ Brace thickness """
        return np.asarray([self.braces[0].profile.T,self.braces[1].profile.T])
    
    @property
    def fy(self):
        """ Brace yield strength """
        return np.asarray([self.braces[0].fy(),self.braces[1].fy()])
    
    def beta(self):
        return (sum(self.b)+ sum(self.h)) / (4*self.b0)
    
    
    def eccentricity(self):
        t = self.angles
        h = self.h
        e0 = 0.5*h[0]/math.sin(math.radians(t[0])) + 0.5*h[1]/math.sin(math.radians(t[1])) + self.gap
        e = e0*math.sin(math.radians(t[0]))*math.sin(math.radians(t[1]))/math.sin(math.radians(sum(t)))-0.5*self.h0
        return e
    
    # Resistance formulas
    def chord_face_failure(self):
                        
        b = self.beta()
        g = self.gamma()                           
        kn = self.eval_KN()
        r = self.strength_reduction()
        fy0 = self.fy0
        if self.chord_face_plate['tp'] == 0:
            t0 = self.t0
        else:
            t0 = self.chord_face_plate['tp']
            
        NRd = []
        NRd0 = r*8.9*kn*fy0*t0**2*math.sqrt(g)*b/gammaM5
        for angle in self.angles:
            s = math.sin(math.radians(angle))            
            NRd.append(NRd0/s)
            
        return np.asarray(NRd)
        
    def chord_shear(self):
        Aeta = self.a_eta()
        # sin of brace angles
        s = np.sin(np.radians(np.array(self.angles)))
        fy0 = self.fy0
        # braces
        NiRd = fy0*Aeta/math.sqrt(3)/s/gammaM5
        # chord
        # Check first, if the chord can sustain the shear force
        if self.V0gap/self.chord.profile.shear_force_resistance() > 1:
            # Chord profile is not sufficient, so set resistance to 0.0.
            #print("Chord is not strong enough for shear forces")
            N0Rd = 0.0

        else:
            N0Rd = fy0*((self.chord.profile.A-Aeta)+Aeta*math.sqrt(1-(self.V0gap/self.chord.profile.VRd)**2))/gammaM5
            
        r = self.strength_reduction()
        NiRd = r*NiRd
        N0Rd = r*N0Rd
        return NiRd, N0Rd
        
    def a_eta(self):
        a = math.sqrt(1/(1+4*self.gap**2/3/self.t0**2))
        return (2*self.h0+a*self.b0)*self.t0
        
    def brace_failure(self):
        r = self.strength_reduction()
        b = np.array(self.b)
        h = np.array(self.h)
        t = np.array(self.t)
        fy = np.array(self.fy)
        fy0 = self.fy0
        if self.chord_face_plate['tp'] == 0:
            t0 = self.t0
        else:
            t0 = self.chord_face_plate['tp']
        beff = np.minimum(10/(self.b0/t0)*fy0*t0/(fy*t),1)*b
        #beff_arr = np.array([min(x,beff[i]) for i, x in enumerate(b)])
            
        NiRd = r*self.fy*t*(2*h-4*t+b+beff)/gammaM5
        return NiRd
        
    def punching_shear(self):
        b = self.b
        fy0 = self.fy0
        if self.chord_face_plate['tp'] == 0:
            t0 = self.t0
        else:
            t0 = self.chord_face_plate['tp']
        bep = np.minimum(10/(self.b0/t0),1)*b 
        r = self.strength_reduction()
        
        s = np.sin(np.radians(np.array(self.angles)))
            
        NRd = r*fy0*t0/math.sqrt(3)/s*(2*self.h/s+b+bep)/gammaM5
        return NRd
    
    def add_chord_face_plate(self,lp=0,bp=0,tp=0):
        """
        Adds a reinforcement plate to the chord face

        Parameters
        ----------
        lp : float
            length of the plate.
        bp : float
            width of the plate.
        tp : float
            thickness of the plate.

        Returns
        -------
        None.

        """
        
        lp_min = 1.5*(self.braces[0].h/np.sin(np.radians(self.angles[0])) + self.gap + 
                                              self.braces[1].h/np.sin(np.radians(self.angles[1])))
        
        lp = max(lp,lp_min)
        
        bp = max(bp,self.b0-2*self.t0)

        tp = max(tp,2*max(self.t))
        
        self.chord_face_plate = {'lp':lp, 'bp':bp,'tp':tp}
        
        
    
    def validity(self,verb=False):
        """ Check the conditions of Table 7.8 regarding the geometry of the joint  
            TODO: i) Gap condition,
                 ii) h0/b0 and hi/bi condition
                iii) bi/ti and hi/ti condition
                    
        
        """
        
        biti_ratio = [True,True]
        hib0_ratio = [True,True]
        b0to_ratio = [True,True]
        gap_ratio = True
        
        b0t0 = self.b0/self.t0
        h0t0 = self.h0/self.t0
        bib0 = self.b/self.b0
        gap_b0 = self.gap/self.b0
        
        bib0_ratio = self.b/self.b0 > max(0.35,0.1+0.01*b0t0)
        
        
        b = self.b
        if self.braces[0].NcEd < 0:
            b1 = b[0]
            b2 = b[1]
        else:
            b1 = b[1]
            b2 = b[0]
                            
        if verb:
            print("  Geometry conditions:")
            print("    a) h0/t0 = {0:4.2f} <= 35".format(h0t0))
            print("    b) b0/t0 = {0:4.2f} <= 35".format(b0t0))
            print("    c) bi/b0 > max(0.35, 0.1 + 0.01*b0/t0) = {0:4.2f}".format(max(0.35,0.1+0.01*b0t0)))
            for i in range(0,2):
                if bib0_ratio[i]:
                    OKNOTOK = "OK!"
                else:
                    OKNOTOK = "NOT OK!"
                print("  Brace {0:.0f}: bi/b0 = {1:4.2f}, {2:s}".format(i,bib0[i],OKNOTOK))
                
            print("  Further conditions for the use of Table 7.10 of EN 1993-1-8:")
            print(" 0.6 <= (b1+b2)/2b1 = {0:4.2f} <= 1.3".format((b1+b2)/2/b1))
            print(" b0/t0 = {0:4.2f} >= 15".format(b0t0))
    
    def info(self):
        """ Prints information on the joint """
        
        print("*** K gap Joint ***")
        
        super().info()
        
        for brace, angle in zip(self.braces,self.angles):
            print("  Brace: {0:s}".format(brace.profile.__repr__()))
            if brace.NcEd < 0:
                Nsense = "compression"
            elif max(brace.ned) > 0:
                Nsense = "tension"
            else:
                Nsense = "no axial force"
            print("  NEd: {0:4.2f} kN ({1:s})".format(brace.NEd*1e-3,Nsense))
            print("    Angle to chord: {0:4.2f} deg".format(angle))
        
        beta = self.beta()
        print("  Beta = {0:4.2f}".format(beta))
        print("  Gamma = {0:4.2f}".format(self.gamma()))        
        print("  Gap = {0:4.2f} mm".format(self.gap))
        print("    t1 + t2 = {0:4.2f} mm".format(sum([brace.profile.t for brace in self.braces])))
        print("    0.5*b0*(1-beta) = {0:4.2f} mm".format(0.5*self.b0*(1-beta)))
        print("    1.5*b0*(1-beta) = {0:4.2f} mm".format(1.5*self.b0*(1-beta)))
        
        print("  Eccentricity = {0:4.2f} mm (Limit = {1:4.2f} mm)".format(self.eccentricity(),0.25*self.h0))
        
        self.validity(True)
        
        print(" -- RESISTANCE -- ")
        
        NRD_chord_face = self.chord_face_failure()
        
        print("    Chord face failure:")
        for i, NRd in enumerate(NRD_chord_face):
            print("      N_{0:.0f}_Rd = {1:4.2f} kN".format(i+1,NRd*1e-3))
            
        
        NiRd, N0Rd = self.chord_shear()
        
        print("    Chord shear failure:")
        for i, NRd in enumerate(NiRd):
            print("      N_{0:.0f}_Rd = {1:4.2f} kN".format(i+1,NRd*1e-3))

        print("      N_0_Rd = {0:4.2f} kN".format(N0Rd*1e-3))
        
        NiRd = self.brace_failure()
        
        print("    Brace failure:")
        for i, NRd in enumerate(NiRd):
            print("      N_{0:.0f}_Rd = {1:4.2f} kN".format(i+1,NRd*1e-3))
            
        NiRd = self.punching_shear()
        
        print("    Punching shear:")
        for i, NRd in enumerate(NiRd):
            print("      N_{0:.0f}_Rd = {1:4.2f} kN".format(i+1,NRd*1e-3))
            
    def draw(self,chord_type="bottom",chord_angle=0,dx=300,dy=250):
        """ Draw the joint 
            input:
                dy .. height of the braces above (below) the chord face
        """
            
        fig, ax = plt.subplots(1)
        
        # draw chord
        super().draw(fig,ax,chord_angle,dx)
        
        """ draw braces
            For the location of the braces, assume
            that the gap is divided equally between the two sides
            of the chord
            
            Assume brace 1 is on the left
        """
        WALL = 1.75
        INNER = 1.25
        CENTER = 1.5
                
        y0 = 0.5*self.h0
        dx0 = 0.5*self.gap
        
        xC = []
        vC = []
        
        for a, brace, K in zip(self.angles,self.braces,[-1,1]):
            arad = math.radians(a)
            
            # Draw the lines of the brace            
            x0 = np.array([K*dx0,y0])
            # Direction vector of the brace
            v = rotate([K,0],K*a)
            vC.append(deepcopy(-v))
            vortho = rotate(v,-K*90)
            
            # First line (inner edge)
            x1 = x0 + dy*v
            ax.plot([x0[0],x1[0]],[x0[1],x1[1]],linewidth=WALL,color='k')
            
            # Second pair of points
            sina = math.sin(arad)
            x0[0] += K * brace.t / sina
            x2 = x1 + brace.t * vortho
            ax.plot([x0[0],x2[0]],[x0[1],x2[1]],linewidth=INNER,color='0.8')
            
            # Center line
            dxin = 0.5*brace.h-brace.t
            x0[0] += K*dxin/sina
            x3 = x2 + dxin*vortho
            xC.append(deepcopy(x0))
            ax.plot([x0[0],x3[0]],[x0[1],x3[1]],linewidth=CENTER,color='k',linestyle='dashdot')
            
            # Second inner wall
            x0[0] += K*dxin/sina
            x4 = x3 + dxin*vortho
            ax.plot([x0[0],x4[0]],[x0[1],x4[1]],linewidth=INNER,color='0.8')
            
            # Second outer wall
            x0[0] += K * brace.t / sina
            x5 = x4 + brace.t / sina * vortho
            ax.plot([x0[0],x5[0]],[x0[1],x5[1]],linewidth=WALL,color='k')
            
            # Plot dashed line to the end of the brace
            ax.plot([x1[0],x5[0]],[x1[1],x5[1]],linewidth=INNER,color='k',linestyle='dashed')
        
        # Find the intersection of the center lines of the braces
        l1 = Line(p1=xC[0],v=vC[0])
        l2 = Line(p1=xC[1],v=vC[1])
        t, p = l1.intersect(l2)
        
        
        # Plot the point and the center lines
        ax.plot(p[0],p[1],'ko')
        for x in xC:
            ax.plot([x[0],p[0]],[x[1],p[1]],linewidth=CENTER,color='k',linestyle='dashdot')
        
        lchord = Line(p1=[0,0],v=[1,0])
        d, p0 = lchord.distance(p)
        
        ax.plot(p0[0],p0[1],'ko')
        ax.plot([p0[0],p[0]],[p0[1],p[1]],linewidth=INNER,color='k',linestyle='dashed')
    
    def design(self):
        """
        Designs the joint

        Returns
        -------
        Utilization ratios N1Ed/NjRd and N2Ed/NjRd

        """
        
        b = self.beta()
        g = self.gamma()
        
        NiEd = self.NEd
        
        N_cff_Rd = self.chord_face_failure()
        N_cs_Rd, N0Rd = self.chord_shear()
        N_bf_Rd = self.brace_failure()
        
        NjRd = np.minimum(N_cff_Rd,N_cs_Rd)
        NjRd = np.minimum(NjRd,N_bf_Rd)
        if b <= 1-1/g:
            N_ps_Rd = self.punching_shear()        
            NjRd = np.minimum(NjRd,N_ps_Rd)

        res = np.array(abs(self.NEd)/NjRd,abs(self.N0/N0Rd))

        return res

class RHSKTGapJoint(RHSJoint):
    """ For KT Gap joints between RHS members """
    
    def __init__(self,chord,braces,angles,gaps=[20,20], **kwargs):
        """ Constructor
            input:
                chord .. chord profile (RHS or SHS class object)
                braces .. list of brace profiles (RHS or SHS class objects)
                angles .. list of angles (degrees) between the braces and the chord
                gap .. gap between the braces
        """
        RHSJoint.__init__(self,chord, **kwargs)
                
        self.braces = braces
        self.gaps = gaps
        self.angles = angles

class RHSYJoint(RHSJoint):
    
    def __init__(self, chord_profile, brace, angle, **kwargs):
        super().__init__(chord_profile, **kwargs)
        
        if not isinstance(brace,SteelMember):
            brace_member = SteelMember(brace,length=1000)
            brace_member.add_section(ned=brace.Ned)
            
            self.brace = brace_member
        else:
                        
            self.brace = brace
        self.angle = angle
        
    @property
    def N1(self):
        """ Force in the brace """
        return self.brace.NEd
    
    @property
    def h(self):
        """ Brace heights """
        return self.brace.profile.H
    
    @property
    def b(self):
        """ Brace widths """
        return self.brace.profile.B
    
    @property
    def t(self):
        """ Brace thickness """
        return self.brace.profile.T
    
    @property
    def fy(self):
        """ Brace yield strength """
        return self.brace.profile.fy
    
    def beta(self):
        return self.b / self.b0
    
    @property
    def eta(self):
        return self.h/self.b0
    
    def fb(self):
        """ Yield strength to be used in chord side wall buckling. """
        
        if self.N1 < 0.0:
            alpha = self.chord.profile.imp_factor[0]
            slend = self.slend()
            phi = 0.5*(1 + alpha*(slend - 0.2) + slend **2)
            chi = 1 / (phi + np.sqrt(phi**2 - slend**2))
            
            fb = chi * self.fy0
            #print(chi,fb)
        else:
            fb = self.fy0
        
        return fb
            
    def chord_face_failure(self):
                        
        b = self.beta()
        if b >= 1:
            # If beta is grater than 1, artificially reduce it to "nearly 1".
            b = 0.99
        kn = self.eval_KN()
        
        r = self.strength_reduction()
        fy0 = self.fy0
        t0 = self.t0
        NRd = r**kn*fy0*t0**2 / ((1-b)*np.sin(np.radians(self.angle)))
        NRd = NRd * (2*self.eta / np.sin(np.radians(self.angle)) + 4*np.sqrt(1-b))
        NRd = NRd/ gammaM5

        return NRd
    
    
    def slend(self):

        sl = 3.46*(self.h0/self.t0 -2)*np.sqrt(1/np.sin(np.radians(self.angle)))
        sl = sl / (np.pi * np.sqrt(E / self.fy0))
        return sl
   
    def chord_web_buckling(self):
        
        fb = self.fb()
        
        kn = self.eval_KN()
        
        r = self.strength_reduction()
        t0 = self.t0
        h = self.h
        NRd = r**kn*fb*t0 / (np.sin(np.radians(self.angle)))
        NRd = NRd * (2*h / np.sin(np.radians(self.angle)) + 10*t0)
        NRd = NRd / gammaM5
        return NRd
    
    def brace_failure(self):
        r = self.strength_reduction()
        b = self.b
        h = self.h
        t = self.t
        fy = self.fy
        fy0 = self.fy0
        t0 = self.t0
        beff = min(10/(self.b0/t0)*fy0*t0/(fy*t),1)*b
            
        NiRd = r*self.fy*t*(2*h-4*t+2*beff)/gammaM5
        return NiRd
    
    def punching_shear(self):
        b = self.b
        fy0 = self.fy0
        t0 = self.t0
        bep = min(10/(self.b0/t0),1.0)*b 
        r = self.strength_reduction()
        
        s = np.sin(np.radians(self.angle))
            
        NRd = r*fy0*t0/math.sqrt(3)/s*(2*self.h/s+2*bep)/gammaM5
        return NRd
    
    def validity(self,verb=False):
        """ Check the conditions of Table 7.8 regarding the geometry of the joint  
            
        """
        
        biti_ratio = True
        hib0_ratio = True
        b0to_ratio = True        
        
        b0t0 = self.b0/self.t0
        h0t0 = self.h0/self.t0
        bib0 = self.b/self.b0        
        
        bib0_ratio = self.b/self.b0 > 0.25
                
        b = self.b                
        if verb:
            
            if bib0_ratio:
                OKNOTOK = "OK!"
            else:
                OKNOTOK = "NOT OK!"
            
            print("  Geometry conditions:")
            print("    a) h0/t0 = {0:4.2f} <= 35".format(h0t0))
            print("    b) b0/t0 = {0:4.2f} <= 35".format(b0t0))
            print("    c) bi/b0 = {0:4.2f} > 0.25, {1:s}".format(bib0,OKNOTOK))            
                        
                
            print("  Further conditions for the use of Table 7.10 of EN 1993-1-8:")
            print(" bi/b0 = {0:4.2f} <= 0.85".format(bib0))
            print(" b0/t0 = {0:4.2f} >= 10".format(b0t0))
    
    def info(self):
        """ Prints information on the joint """
        
        print("*** Y Joint ***")
        
        super().info()
        
        print("  Brace: {0:s}".format(self.brace.profile.__repr__()))
        if self.brace.NEd < 0:
            Nsense = "compression"
        elif self.brace.Ned > 0:
            Nsense = "tension"
        else:
            Nsense = "no axial force"
        print("  NEd: {0:4.2f} kN ({1:s})".format(self.brace.NEd*1e-3,Nsense))
        print("    Angle to chord: {0:4.2f} deg".format(self.angle))
        
        beta = self.beta()
        print("  Beta = {0:4.2f}".format(beta))
        print("  Gamma = {0:4.2f}".format(self.gamma()))                        
        
        self.validity(True)
        
        print(" -- RESISTANCE -- ")
        
        NRD_chord_face = self.chord_face_failure()
        
        print("    Chord face failure:")
        print("      N_1_Rd = {0:4.2f} kN".format(NRD_chord_face*1e-3))
        
        NiRd = self.chord_web_buckling()
        
        print("    Chord side wall buckling:")        
        print("      N_1_Rd = {0:4.2f} kN".format(NiRd*1e-3))        
        
        NiRd = self.brace_failure()
        
        print("    Brace failure:")        
        print("      N_1_Rd = {0:4.2f} kN".format(NiRd*1e-3))
            
        NiRd = self.punching_shear()
        
        print("    Punching shear:")        
        print("      N_1_Rd = {0:4.2f} kN".format(NiRd*1e-3))
    
    def design(self):
        """
        Designs the joint

        Returns
        -------
        Utilization ratio N1Ed/NjRd.

        """
        
        b = self.beta()
        g = self.gamma()
        
        
        if b < 0.85:
            NjRd = self.chord_face_failure()
        
        else:
            N_cff_Rd = self.chord_face_failure()
            N_csw_Rd = self.chord_web_buckling()
            N_bf_Rd = self.brace_failure()
            
            # Linear interpolation between chord face failure at b = 0.85
            # and chord side wall failure at b = 1.0.
            # The denominator is 1-0.85 = 0.15.            
            N_interp_Rd = N_cff_Rd + (b-0.85)/0.15*(N_csw_Rd-N_cff_Rd)
            
            NjRd = min(N_interp_Rd,N_bf_Rd)
            
            if b <= 1-1/g:
                N_ps_Rd = self.punching_shear()                
                NjRd = min(NjRd,N_ps_Rd)

        res = abs(self.N1)/NjRd        

        return res
            
            
    
class RHSXJoint(RHSYJoint):
    
    def __init__(self, chord_profile, brace, angle, **kwargs):
        super().__init__(chord_profile, brace, angle, **kwargs)
    
    def fb(self):
        """ Yield strength to be used in chord side wall buckling. """        
                
                
        if self.N1 < 0.0:            
            alpha = self.chord.imp_factor[0]
            slend = self.slend()
            phi = 0.5*(1 + alpha*(slend - 0.2) + slend **2)
            chi = 1 / (phi + np.sqrt(phi**2 - slend**2))
            
            res = 0.8*chi * self.fy0*np.sin(np.radians(self.angle))            
        else:
            res = self.fy0
        
        return res
    
    def chord_web_buckling(self):
        
        # Using the chord web buckling of the Y joint class, but
        # with fb calculated for X joints.
        NRd = super().chord_web_buckling()
        
        if np.cos(np.radians(self.angle)) > self.h/self.h0:
            # Take the minimium of chord web buckling for X joints
            # and chord shear for K and N joints
            # 
            Nshear = self.fy0*self.chord.Ashear/np.sqrt(3)/np.sin(np.radians(self.angle))
            # The strength reduction factor is already taken into account
            # for the chord web buckling of the Y joint class method.
            # It must be applied for the chord shear.
            r = self.strength_reduction()
            NRd = min(NRd,r*Nshear)
        
        NRd = NRd / gammaM5
        return NRd
        
    def info(self):
        """ Prints information on the joint """
        
        print("*** X Joint ***")
        
        RHSJoint.info(self)
        
        print("  Brace: {0:s}".format(self.brace.__repr__()))
        if self.brace.Ned < 0:
            Nsense = "compression"
        elif self.brace.Ned > 0:
            Nsense = "tension"
        else:
            Nsense = "no axial force"
        print("  NEd: {0:4.2f} kN ({1:s})".format(self.brace.Ned*1e-3,Nsense))
        print("    Angle to chord: {0:4.2f} deg".format(self.angle))
        
        beta = self.beta()
        print("  Beta = {0:4.2f}".format(beta))
        print("  Gamma = {0:4.2f}".format(self.gamma()))                        
        
        self.validity(True)
        
        print(" -- RESISTANCE -- ")
        
        NRD_chord_face = self.chord_face_failure()
        
        print("    Chord face failure:")
        print("      N_1_Rd = {0:4.2f} kN".format(NRD_chord_face*1e-3))
        
        NiRd = self.chord_web_buckling()
        
        print("    Chord side wall buckling:")        
        print("      N_1_Rd = {0:4.2f} kN".format(NiRd*1e-3))        
        
        NiRd = self.brace_failure()
        
        print("    Brace failure:")        
        print("      N_1_Rd = {0:4.2f} kN".format(NiRd*1e-3))
            
        NiRd = self.punching_shear()
        
        print("    Punching shear:")        
        print("      N_1_Rd = {0:4.2f} kN".format(NiRd*1e-3))

def Y_example():
    
    
    chord = SHS(140,6,fy=700)

    brace = SHS(140,8,fy=700)
    brace.Ned = -50e3
    
    
    YJoint = RHSYJoint(chord,brace,45,N0=-887.5e3)

    YJoint.r = 1.0
    
    YJoint.info()
    
    return YJoint

def K_example():
    
    chord = RHS(80,140,6,fy=700)
    b1 = RHS(70,70,3,fy=700)
    b2 = RHS(40,60,4,fy=700)
    
    b1.Ned = 123.743e3
    b2.Ned = -123.743e3
    
    Kjoint = RHSKGapJoint(chord, [b1,b2], [45,45],40,N0=-362500.0)
    
    print(Kjoint.eval_KN())
    
    return Kjoint
    
if __name__ == '__main__':
    
    from sections.steel.RHS import RHS, SHS
    #YJ = Y_example()
    
    #KJ = K_example()
    
    """
    C = RHS(60,40,3,fy=700)
    C.material.fy = 770
    B1 = RHS(40,40,4,fy=700)
    B1.material.fy = 770
    B2 = RHS(40,40,4,fy=700)
    B2.material.fy = 770
    """
    
    C = RHS(60,40,3,fy=700)
    C.material.fy = 770
    B1 = RHS(40,40,4,fy=700)
    B1.material.fy = 770
    B2 = RHS(40,40,4,fy=700)
    B2.material.fy = 770
    
    C.Ned = -500e3
    B1.Ned = 200e3
    B2.Ned = -200e3
    
    
    
    K = RHSKGapJoint(C, [B1,B2], [45,45],15,N0=C.Ned)
    
    K.info()
    
    """
    
    
    
    
    
    chord = SHS(200,8,fy=420)
    
    chord.Ned = -1364e3
    b1 = SHS(150,6,fy=420)
    b2 = SHS(150,6,fy=420)
        
    
    J = RHSKGapJoint(chord,[b1,b2], [45,45], 28)
    
    J.N0 = -1364e3
    J.braces[0].Ned = 600e3
    J.braces[1].Ned = -600e3
    
    #print(J.N0gap*1e-3)
    #print(J.V0gap*1e-3)
    
    #J.info()
    #J.draw()
    
    b1.Ned = -600e3
    
    #X = RHSXJoint(chord,b1,45)
    
    #X.chord_web_buckling()
    
    #X.info()
    
    #l1 = Line(p1=[0,0],v=[1,0])
    #l2 = Line(p1=[1,1],v=[0,1])
    
    #t, x0 = l1.intersect(l2)
    
    # Y Joint example from SSAB's book  (Ex. 3.1) 
    chord = SHS(200,8,fy=420)
    brace = SHS(100,5,fy=420)
    
    brace.Ned = 590e3
    
    Y = RHSYJoint(chord,brace,45,N0=-936e3)
    
    # Y Joint example from SSAB's book  (Chapter 7) 
    chord = SHS(200,8,fy=420)
    brace = SHS(120,5,fy=420)
    
    brace.Ned = 750e3
    
    Y = RHSYJoint(chord,brace,54,N0=-527e3,M0=-653e3*200)
    
    Y.info()
    """


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Wed Mar 21 13:58:54 2018

Class for modelling cold-formed sections, according to 

EN 1993-1-3, Annex C

@author: kmela
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches

class OpenProf:
    """
    OpenProf Class for general open cold-formed profiles
    Main purpose is to calculate cross-sectional properties as
    described in EN 1993-1-3, Annex C.
    """

    def __init__(self,nodes,t):
        """ Constructor
            input:
                nodes .. list of nodal coordinates
                    nodes[i] = [y_i, z_i]
                t .. array of wall thicknesses or a single value
        """        
        n = len(nodes)
        
        self.nodes = []
        for node in nodes:
            if isinstance(node,Node):
                self.nodes.append(node)
            else:
                self.nodes.append(Node(y=node[0],z=node[1]))
        
        #self.nodes = np.array(nodes)
        if isinstance(t,float):
            self.t = t*np.ones(n-1)
        else:
            self.t = np.array(t)
                        
        self.segments = []
        
        for i in range(1,n):                      
            self.add_segment(self.nodes[i-1],self.nodes[i],self.t[i-1])
            #self.add_segment(self.nodes[i,:],self.nodes[i-1,:],self.t[i-1])

    @property
    def n(self):
        """ Number of nodes """
        return len(self.nodes)
    
    def y_coord(self):
        """ Return y coordinates """
        return np.array([node.y for node in self.nodes])
        #return self.nodes[:,0]
    
    def z_coord(self):
        """ Return z coordinates """
        return np.array([node.z for node in self.nodes])
        #return self.nodes[:,1]
    
    def add_segment(self,n1,n2,t):
        """ Adds new line segment to cross section """
        new_segment = Segment(n1,n2,t)
        self.segments.append(new_segment)
                  
    def draw(self,node_labels=False,seg_labels=False,axes_on=True,
             sc_on=True,gc_on=True,coord_axes=False):
        
        fig, ax = plt.subplots(1)         
        """ draw nodes """
        for i, node in enumerate(self.nodes):
            if node_labels:
                node.draw(str(i),ax)
            else:
                node.draw(axes=ax)
            #plt.plot(node[0],node[1],'ro')            
            #plt.text(node[0],node[1],str(i))
        
        #for i in range(self.n):
        #    plt.text(self.nodes[i][0],self.nodes[i][1],str(i))
            
        
        """ draw line segments """
        for i, s in enumerate(self.segments):
            if seg_labels:
                s.draw(str(i),axes=ax)
            else:
                s.draw(axes=ax)
            #X = s.nodes
            #plt.plot(X[:,0],X[:,1],'b')
            #Xmid = X[0,:]+0.5*(X[1,:]-X[0,:])
            #plt.text(Xmid[0],Xmid[1],str(i))
                
    
        if gc_on:
            ygc, zgc, A = self.centroid()        
            ax.plot(ygc,zgc,'+r')
        
        if sc_on:
            ysc, zsc = self.shear_center()        
            ax.plot(ysc,zsc,'or')
        
        ax.axis('equal')
        
        if axes_on is False:
            plt.axis('off')
        #ax.show()        

        if coord_axes:
            x = self.y_coord()
            y = self.z_coord()
            dx = 0.5*(np.max(x)-np.min(x))
            dy = 0.5*(np.max(y)-np.min(y))
            hw = 3
            ax.arrow(0,0,min(dx,dy),0,head_width=hw,facecolor='k')
            ax.text(min(dx,dy),0.05*dy,'y',verticalalignment='bottom')
            ax.arrow(0,0,0,min(dx,dy),head_width=hw,facecolor='k')
            ax.text(-0.05*dy,min(dx,dy),'z',horizontalalignment='right')

        return fig

    def area(self):        
        """ Cross-sectional area """
        A = 0.0
            
        for s in self.segments:
            A += s.area()        
        return A
                
    def first_moments(self):
        """ First moments of area  """                
        Sy0 = 0.0        
        Sz0 = 0.0
        for s in self.segments:            
            Sy0 += s.first_moment_y()
            Sz0 += s.first_moment_z()
        return Sy0, Sz0
            
    def centroid(self):
        """ Centroid of area """
        Sy0, Sz0 = self.first_moments()
        A = self.area()
        zgc = Sy0/A
        ygc = Sz0/A
            
        return ygc, zgc, A
    
    def second_moments(self):
        """ Second moments of area """                
        Iy0 = 0.0        
        Iz0 = 0.0
        for s in self.segments:            
            Iy0 += s.second_moment_y()
            Iz0 += s.second_moment_z()
        return Iy0, Iz0
    
    def second_moments_gc(self):
        """ Second moments of area through the center of gravity """
        ygc, zgc, A = self.centroid()
        Iy0, Iz0 = self.second_moments()
        Iy = Iy0-A*zgc**2
        Iz = Iz0-A*ygc**2
        return Iy, Iz
    
    def Iy0(self):
        """ Second moment of area with respect to Y-axis """
        Iy0 = 0.0        
        for s in self.segments:            
            Iy0 += s.second_moment_y()           
        return Iy0
    
    def Iz0(self):
        """ Second moment of area with respect to Z-axis """
        Iz0 = 0.0        
        for s in self.segments:            
            Iz0 += s.second_moment_z()           
        return Iz0
    
    def Iy(self):
        """ Second moment of area with respect to centroid """
        ygc, zgc, A = self.centroid()
        Iy0 = self.Iy0()
        return Iy0-A*zgc**2
    
    def Iz(self):
        """ Second moment of area with respect to centroid """
        ygc, zgc, A = self.centroid()
        Iz0 = self.Iz0()
        return Iz0-A*ygc**2
    
    def product_moment(self):
        """ Product moment with respect to original y- and z-axes """
        Iyz0 = 0.0
        for s in self.segments:            
            Iyz0 += s.product_moment()
            
        return Iyz0
    
    def product_moment_gc(self):
        """ Product moment with respect to center of gravity """
        A = self.area()
        Sy0, Sz0 = self.first_moments()
        Iyz0 = self.product_moment()
        return Iyz0-Sy0*Sz0/A
    
    def sectorial_coordinates(self):
        """ Determine sectorial coordinates """
        n = self.n
        y = self.y_coord()
        z = self.z_coord()
        w = np.zeros(n)
        #print(y,z)
        for i in range(n-1):
            w[i+1] = w[i] + y[i]*z[i+1]-y[i+1]*z[i]
            
        return w
    
    def principal_axes(self):
        """ Principal axes """
        Iy, Iz = self.second_moments_gc()
        Iyz = self.product_moment_gc()
        d = Iz-Iy
        if abs(d) > 1e-8:
            a = .5*np.atan(2*Iyz/d)
        else:
            a = 0
            
        I1 = 0.5*(Iy+Iz+np.sqrt(d^2+4*Iyz^2))
        I2 = .5*(Iy+Iz-np.sqrt(d^2+4*Iyz^2))
        
        return [I1,I2], a
    
    def mean_sectorial_moment(self):
        n = self.n
        dA = [s.area() for s in self.segments]
        Iw = 0
        w = self.sectorial_coordinates()
        
        for i in range(n-1):
            Iw += 0.5*dA[i]*(w[i+1]+w[i])
            
        return Iw
    
        
    def mean_sectorial_coordinate(self):
        Iw = self.mean_sectorial_moment()
        A = self.area()
        return Iw/A
    
    def sectorial_constants(self):
        dA = [s.area() for s in self.segments]
        Iyw0 = 0.0
        Izw0 = 0.0
        Iww0 = 0.0
        
        Iw = self.mean_sectorial_moment()
        A = self.area()
        Sy0, Sz0 = self.first_moments()
        
        y = self.y_coord()
        z = self.z_coord()
        w = self.sectorial_coordinates()
        
        for i in range(self.n-1):
            Iyw0 += dA[i]*(2*y[i]*w[i]+2*y[i+1]*w[i+1]+y[i]*w[i+1]+y[i+1]*w[i])/6
            Izw0 += dA[i]*(2*z[i]*w[i]+2*z[i+1]*w[i+1]+z[i]*w[i+1]+z[i+1]*w[i])/6
            Iww0 += dA[i]*(w[i+1]**2+w[i]**2+w[i+1]*w[i])/3                    
    
        Iyw = Iyw0 - Sz0*Iw/A
        Izw = Izw0 - Sy0*Iw/A
        Iww = Iww0 - Iw**2/A
        
        return Iyw, Izw, Iww
    
    def shear_center(self):
        
        Iyw, Izw, Iww = self.sectorial_constants()
        Iy, Iz = self.second_moments_gc()
        Iyz = self.product_moment_gc()
        
        d = Iy*Iz-Iyz**2
        
        if d > 1e-8:
            y_sc = (Izw*Iz-Iyw*Iyz)/d
            z_sc = (-Iyw*Iy+Izw*Iyz)/d
        else:
            y_sc = 0.0
            z_sc = 0.0
        
        return y_sc, z_sc
    
    def warping_constant(self):
        
        y_sc, z_sc = self.shear_center()        
        Iyw, Izw, Iww = self.sectorial_constants()
        
        return Iww + z_sc*Iyw - y_sc*Izw
    
    def torsion_constant(self):
        dA = np.array([s.area() for s in self.segments])
        
        return sum(dA*np.array(self.t)**2/3)
    
    def torsion_modulus(self):
        
        return self.torsion_constant()/min(self.t)
    
    def split_segment(self,n,s):
        """ Splits a segment into two by adding a point 
            input:
                n .. index of the segment
                s .. local coordinate along the segment, where
                     the new point is to be added        
        """
        
        seg = self.segments[n]
        
        # Get nodal coordinates
        nodal_coords = seg.coord        
        new_point = nodal_coords[0] + s*(nodal_coords[1]-nodal_coords[0])    
        new_node = Node(new_point[0],new_point[1])
        
        #print(seg.nodes[0].y,seg.nodes[0].z)
        
        new_seg1 = Segment(seg.nodes[0],new_node,seg.t)
        new_seg2 = Segment(new_node,seg.nodes[1],seg.t)
        
        
        self.nodes.insert(n+1,new_node)
        self.segments.remove(seg)
        self.segments.insert(n,new_seg1)
        self.segments.insert(n+1,new_seg2)
    
    def export_cufsm(self):
        """ Write data to be exported to cufsm in Matlab """
                
        """ Write nodes """
        print("Nodes:")
        for i, node in enumerate(self.nodes):        
            print('{0:3g} {1:6.4f} {2:6.4f} 1 1 1 1 1.000'.format(i+1,node.y,node.z))

        """ Write elements """
        print("Elements:")
        for i, seg in enumerate(self.segments):        
            print('{0:3g} {1:6.4f} {2:6.4f} {3:6.5f} 100'.format(i + 1, i + 1, i + 2, seg.t))

class Node:
    """ Node of an open section profile """
    
    def __init__(self,y,z):
        """ Constructor 
            y is the horizontal axis and
            z is the vertical axis
        """
        
        self.y = y
        self.z = z
    
    def __repr__(self):
        
        return "y = {0:4.3f}, z = {1:4.3f}".format(self.y,self.z)
    
    @property
    def coord(self):
        """ Return numpy array of coordinates """
        return np.array([self.y,self.z])
    
    def draw(self,label=None,axes=None):        
        
        if axes is None:
            plt.plot(self.y,self.z,'ok',markersize=4)
            if label is not None:
                plt.text(self.y,self.z,label)
        else:
            axes.plot(self.y,self.z,'ok',markersize=4)
            if label is not None:
                axes.text(self.y,self.z,label)
        
        
        
class Segment:
    """ Line segment that is a part of a cross section """
    
    def __init__(self,n1,n2,t):
        """ Constructor
            input:
                n1 .. node 1 (list of coordinates)
                n2 .. node 2 (list of coordinates)
                t .. thickness
        """
        
        """ nodes is a numpy array with nodal coordinates
            nodes[:,0] .. y coordinates
            nodes[:,1] .. z coordinates
        """
        self._n1 = n1
        self._n2 = n2
        self.nodes = np.array([n1,n2])
        self.t = t
        
    def __repr__(self):
        
        n1 = 'n1 = [{0:4.2f}, {1:4.2f} '.format(self._n1.y,self._n1.z)
        n2 = 'n2 = [{0:4.2f}, {1:4.2f}] '.format(self._n2.y,self._n2.z)
        
        return n1 + n2 + " t = {0:4.2f}".format(self.t)
    
    @property
    def coord(self):
        """ Nodal coordinates """
        return np.array([self._n1.coord, self._n2.coord])
                   
    @property
    def y(self):
        """ Y-coordinates of the segment """
        return np.array([self._n1.y,self._n2.y])
    
    @property
    def z(self):
        """ Z-coordinates of the segment """
        return np.array([self._n1.z,self._n2.z])
    
    def mid_point(self):
        """ coordinates of the middle point of the segment """
        
        return self._n1.coord + 0.5*(self._n2.coord-self._n1.coord)
    
    
    def length(self):
        """ Length """
        return np.linalg.norm(self._n2.coord-self._n1.coord)
        #return np.sqrt(sum((self.nodes[1,:]-self.nodes[0,:])**2))

    def area(self):
        """ Cross-sectional area """
        return self.t*self.length()
        #return self.t*np.sqrt(sum((self.nodes[1,:]-self.nodes[0,:])**2))
    
    def first_moment_y(self):
        """ First moment of area with respect to y axis"""
        dA = self.area()
        return sum(self.z)*dA*0.5
        #return sum(self.nodes[:,1])*dA*0.5
    
    def second_moment_y(self):
        """ Second moment of area with respect to y axis"""
        dA = self.area()
        z = self.z
        return (z[1]**2+z[0]**2+z[1]*z[0])*dA/3
        #return (sum(self.nodes[:,1]**2)+self.nodes[0,1]*self.nodes[1,1])*dA/3
    
    def first_moment_z(self):
        """ First moment of area with respect to z axis"""
        dA = self.area()
        return sum(self.y)*dA*0.5
        #return sum(self.nodes[:,0])*dA*0.5
    
    def second_moment_z(self):
        """ Second moment of area with respect to z axis"""
        dA = self.area()
        y = self.y
        return (y[1]**2+y[0]**2+y[1]*y[0])*dA/3
        #return (sum(self.nodes[:,0]**2)+self.nodes[0,0]*self.nodes[1,0])*dA/3
    
    def product_moment(self):
        """ Product moment of areaq with respect to original y- and z-axes """
        dA = self.area()
        y = self.y
        z = self.z
        Iyz0 = (2*y[0]*z[0] + 2*y[1]*z[1] + y[0]*z[1] + y[1]*z[0])*dA/6
        #Iyz0 = (sum(2*self.nodes.prod(1))+self.nodes[0,0]*self.nodes[1,1]+\
        #   self.nodes[0,1]*self.nodes[1,0])*dA/6
        return Iyz0
    
    def draw(self,label=None,axes=None):
        """ Draw segment """
        
        if axes is None:
            plt.plot(self.y,self.z,'b')
            if label is not None:            
                mid = self.mid_point()
                plt.text(mid[0],mid[1],label)
        else:
            lw0 = 3
            
            axes.plot(self.y,self.z,linewidth=self.t*lw0,color='b')
            
            #xy = (self.y[0],self.z[0])        
            #seg1 = patches.Rectangle(xy,width=self.length(),height=0.5*self.t,fill=False)            
            #seg2 = patches.Rectangle(xy,width=self.length(),height=0.5*self.t,fill=False)            
            #axes.add_patch(seg)
        
            if label is not None:     
                if self.t > 0.0:
                    mid = self.mid_point()
                    axes.text(mid[0],mid[1],label)        


if __name__ == "__main__":
    
    n = []
    n.append([20,3])
    n.append([20,0])
    n.append([0,0])
    n.append([0,60])
    n.append([20,60])
    n.append([20,57])
    
    C = OpenProf(n,1.5)
    C.draw()
    


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Fri Nov 27 08:20:46 2020

Data for loads and load combinations

EN 1990 and EN 1991

@author: kmela, Viktor Haimi
"""

import numpy as np
import copy
from abc import ABC, abstractmethod

# RIL 201-1-2008 Osa 0 Taulukko A1.1
# Combination factors for live loads.
comb_factors = {
    'A': [0.7, 0.5, 0.3],
    'B': [0.7, 0.5, 0.3],
    'C': [0.7, 0.7, 0.3],
    'D': [0.7, 0.7, 0.6],
    'E': [1.0, 0.9, 0.8],
}

# RIL 205-2-2009 Taulukko 2.2
comb_factors_fire = {
    'A': [0.3, 0.3],
    'B': [0.3, 0.3],
    'C': [0.3, 0.3],
    'D': [0.6, 0.6],
    'E': [0.8, 0.8],
    'F': [0.6, 0.6],
}


class Load:
    """ General class for loads """
    
    def __init__(self,value=1.0,comb_factors=[1.0,1.0,1.0]):
        """ Constructor """
        
        self.value = value
        self.comb_factors = comb_factors
        self.final_value = None
        
    def combined(self,combi=0):
        """ Returns the combined value
            :param combi: integer (0, 1, 2) for choosing the combination factor
        """
        return self.value*self.comb_factors[combi]

    def __copy__(self):
        return Load(self.value, self.comb_factors)

    def __repr__(self):
        return f"Load"

class AreaLoad(Load):
    def __init__(self, value=1.0, comb_factors=[1.0, 1.0, 1.0]):
        super().__init__(value, comb_factors)

    def __copy__(self):
        return AreaLoad(self.value, self.comb_factors)

    def __repr__(self):
        return "Area Load"

class ImposedLoad(Load):
    """ Class for imposed (live) loads """
    
    def __init__(self,value=2.0,comb_factors=[0.7,0.5,0.2]):
        """ Constructor
        
            :param value_ground: value of snow load on ground [kN/m2]
        """
        
        super().__init__(value,comb_factors)

    def __copy__(self):
        return ImposedLoad(self.value, self.comb_factors)

    def __repr__(self):
        return f"Imposed Load"


class SnowLoad(Load):
    """ Class for snow loads """
    
    def __init__(self,value_ground=2.0,comb_factors=[0.7,0.5,0.2]):
        """ Constructor
        
            :param value_ground: value of snow load on ground [kN/m2]
        """
        
        if value_ground < 2.75:
            comb_factors = [0.7,0.4,0.2]
        else:
            comb_factors = [0.7,0.5,0.2]
            
        super().__init__(value_ground,comb_factors)

    def __copy__(self):
        return SnowLoad(self.value, self.comb_factors)

    def __repr__(self):
        
        return f"Snow load: sk = {self.value:.1f} kN/m2"


class WindLoad(Load):
    """ Class for wind loads """
    
    def __init__(self,value=0.6,comb_factors=[0.6,0.2,0.0]):
        """ Constructor
        
            :param value: basic value of wind load [kN/m2]
        """
            
        super().__init__(value,comb_factors)

    def __copy__(self):
        return WindLoad(self.value, self.comb_factors)

    def __repr__(self):
        
        return "Wind load"


class Combiner(ABC):
    def __init__(self, structure_class='E', CC=2, kk_jako=1):
        """

        @param structure_class: Rakennustyyppi 'A' - 'E'
        @param CC: seuraamusluokka 1 - 3, käytetään vain ULSCombinerissa
        @param kk_jako: kehien välinen etäisyys metreissä
        """
        self.structure_class = structure_class
        self.CC = CC
        if CC == 3:
            self.kfi = 1.1
        elif CC == 2:
            self.kfi = 1.0
        elif CC == 1:
            self.kfi = 0.9
        self.kk_jako = kk_jako
        
        self.G_k = [] # List of permanent loads
        self.Q_k = [] # List of variable loads

    def get_comb_factors(self):
        """ Combination factors for live load """
        return comb_factors[self.structure_class]

    def add(self, *args):
        """ Adds a load to the load combination """
        
        for l in args:
            # l[0] .. Load class object
            # l[1] .. 'x', 'y', or 'z', direction of the load
            # l[2] .. ???
            load = [copy.copy(l[0]), l[1], l[2]]
            if isinstance(load[0], (SnowLoad, WindLoad, ImposedLoad)):
                if isinstance(load[0], ImposedLoad):
                    load[0].comb_factors = self.get_comb_factors()
                # Change the value of the load to a line load
                load[0].value *= self.kk_jako
                self.Q_k.append(load)
            elif isinstance(load[0], (Load, AreaLoad)):
                if isinstance(load[0], AreaLoad):
                    # Change the value of the load to a line load
                    load[0].value *= self.kk_jako
                    
                load[0].comb_factors = self.get_comb_factors()
                self.G_k.append(load)

    def sum_G_k(self):
        s = sum([G[0].value for G in self.G_k])
        return s

    def sum_Q_k(self, kahveliID):
        s = sum([Q[0].combined(kahveliID) for Q in self.Q_k])
        return s

    def sum_Q_k_excluding(self, exclude_load, kahveliID):
        mod = self.Q_k.copy()
        filtered = filter(lambda q: q[0] != exclude_load, mod)
        s = sum([Q[0].combined(kahveliID) for Q in filtered])
        return s

    def use_reduction_factor(self, length):
        """
        Reduce Imposed load by a factor alpha_A

        @param length: In meters
        """
        if self.structure_class in ('A', 'B', 'C', 'D'):
            for load in self.Q_k:
                if isinstance(load[0], ImposedLoad):
                    alpha = (5/7) * load[0].comb_factors[0] + 10 / (self.kk_jako * length)
                    alpha = np.clip(alpha, 0.7, 1.0)
                    load[0].value *= alpha

    @staticmethod
    def get_single_result(res):
        result = []
        if 'G' in res['combination']:
            for load in res['combination']['G']:
                load['load'][0].final_value = load['load'][0].value * load['factor']
            result.extend(list(map(lambda x: x['load'], res['combination']['G'])))
        if 'Q1' in res['combination']:
            for load in res['combination']['Q1']:
                load['load'][0].final_value = load['load'][0].value * load['factor']
            result.extend(list(map(lambda x: x['load'], res['combination']['Q1'])))
        if 'Qi' in res['combination']:
            for load in res['combination']['Qi']:
                load['load'][0].final_value = load['load'][0].value * load['factor']
            result.extend(list(map(lambda x: x['load'], res['combination']['Qi'])))


        max_load_id = None
        for load in result:
            if max_load_id is None:
                max_load_id = max(load[2])
            if max(load[2]) > max_load_id:
                max_load_id = max(load[2])

        loads_by_id = [None] * (max_load_id + 1)
        qval_y = [None] * (max_load_id + 1)
        qval_x = [None] * (max_load_id + 1)
        for load in result:
            for id in load[2]:
                if loads_by_id[id] is None:
                    loads_by_id[id] = [load]
                else:
                    loads_by_id[id].append(load)
        for i, loads in enumerate(loads_by_id):
            if loads is None:
                continue
            else:
                qx = 0
                qy = 0
                for load in loads:
                    if load[1] == 'x':
                        qx += round(load[0].final_value, 2)
                    elif load[1] == 'y':
                        qy += round(load[0].final_value, 2)
                qval_y[i] = qy
                qval_x[i] = qx
        return {'x': qval_x, 'y': qval_y}

    @staticmethod
    def get_single_split_result(res):
        def partial_result(part):
            result = []
            for load in res['combination'][part]:
                load['load'][0].final_value = load['load'][0].value * load['factor']
            result.extend(list(map(lambda x: x['load'], res['combination'][part])))

            max_load_id = None
            for load in result:
                if max_load_id is None:
                    max_load_id = max(load[2])
                if max(load[2]) > max_load_id:
                    max_load_id = max(load[2])

            loads_by_id = [None] * (max_load_id + 1)
            qval_y = [None] * (max_load_id + 1)
            qval_x = [None] * (max_load_id + 1)

            for load in result:
                for id in load[2]:
                    if loads_by_id[id] is None:
                        loads_by_id[id] = [load]
                    else:
                        loads_by_id[id].append(load)
            for i, loads in enumerate(loads_by_id):
                if loads is None:
                    continue
                else:
                    qx = 0
                    qy = 0
                    for load in loads:
                        if load[1] == 'x':
                            qx += round(load[0].final_value, 2)
                        elif load[1] == 'y':
                            qy += round(load[0].final_value, 2)
                    qval_y[i] = qy
                    qval_x[i] = qx
            return {'x': qval_x, 'y': qval_y}

        split_result = {}
        if 'G' in res['combination']:
            split_result['G'] = partial_result('G')
        if 'Q1' in res['combination']:
            split_result['Q1'] = partial_result('Q1')
        if 'Qi' in res['combination']:
            split_result['Qi'] = partial_result('Qi')
        return split_result

    @abstractmethod
    def combine(self):
        pass

    @abstractmethod
    def get_result(self):
        pass


class ULSCombiner(Combiner):
    def combine(self):
        vals = []
        case1 = []
        case3 = []
        for l in self.G_k:
            case1.append({'load': l, 'factor': 1.35 * self.kfi})
            case3.append({'load': l, 'factor': 1.15 * self.kfi})

        vals.append({'value': 1.35 * self.kfi * self.sum_G_k(), 'combination': {'G': case1}})

        for l in self.Q_k:
            case2q1 = []
            case2qi = []
            case2q1.append({'load': l, 'factor': 1.5 * self.kfi})
            for lo in self.Q_k:
                if l[0] is not lo[0]:
                    case2qi.append({'load': lo, 'factor': 1.5 * self.kfi * lo[0].comb_factors[0]})

            vals.append({'value': 1.15 * self.kfi * self.sum_G_k() + 1.5 * self.kfi * l[0].value +
                                  1.5 * self.kfi * self.sum_Q_k_excluding(l[0], 0), 'combination': {'G': case3, 'Q1': case2q1, 'Qi': case2qi}})

        return max(vals, key=lambda x: x['value'])

    def get_result(self):
        """

        @return: returns the combination with the highest value
        """
        if len(self.Q_k) + len(self.G_k) <= 0:
            raise Exception('Give at least one Load with add method')
        res = self.combine()
        return self.get_single_result(res)


class SLSCombiner(Combiner):
    def combine(self):
        characteristic_combination = []
        frequent_combination = []
        quasi_permanent_combination = []
        case1 = []
        for l in self.G_k:
            case1.append({'load': l, 'factor': 1})

        if len(self.Q_k) == 0:
            characteristic_combination.append({'value': self.sum_G_k(),
                                               'combination': {'G': case1}})
            frequent_combination.append({'value': self.sum_G_k(),
                                         'combination': {'G': case1}})
        case4 = []
        for l in self.Q_k:
            case2q1 = []
            case2qi = []
            case3q1 = []
            case3qi = []
            case2q1.append({'load': l, 'factor': 1})
            case3q1.append({'load': l, 'factor': l[0].comb_factors[1]})
            case4.append({'load': l, 'factor': l[0].comb_factors[2]})

            for lo in self.Q_k:
                if l is not lo:
                    case2qi.append({'load': lo, 'factor': lo[0].comb_factors[0]})
                    case3qi.append({'load': lo, 'factor': lo[0].comb_factors[2]})
            characteristic_combination.append({'value': self.sum_G_k() + l[0].value + self.sum_Q_k_excluding(l[0], 2),
                                               'combination': {'G': case1, 'Q1': case2q1, 'Qi': case2qi}})
            frequent_combination.append({'value': self.sum_G_k() + l[0].combined(1) + self.sum_Q_k_excluding(l[0], 2),
                                         'combination': {'G': case1, 'Q1': case3q1, 'Qi': case3qi}})

        quasi_permanent_combination.append({'value': self.sum_G_k() + self.sum_Q_k(2),
                                            'combination': {'G': case1, 'Qi': case4}})

        for cc in characteristic_combination:
            if 'G' in cc['combination']:
                if len(cc['combination']['G']) == 0:
                    cc['combination'].pop('G')
            if 'Q1' in cc['combination']:
                if len(cc['combination']['Q1']) == 0:
                    cc['combination'].pop('Q1')
            if 'Qi' in cc['combination']:
                if len(cc['combination']['Qi']) == 0:
                    cc['combination'].pop('Qi')

        for fc in frequent_combination:
            if 'G' in fc['combination']:
                if len(fc['combination']['G']) == 0:
                    fc['combination'].pop('G')
            if 'Q1' in fc['combination']:
                if len(fc['combination']['Q1']) == 0:
                    fc['combination'].pop('Q1')
            if 'Qi' in fc['combination']:
                if len(fc['combination']['Qi']) == 0:
                    fc['combination'].pop('Qi')

        for qc in quasi_permanent_combination:
            if 'G' in qc['combination']:
                if len(qc['combination']['G']) == 0:
                    qc['combination'].pop('G')
            if 'Qi' in qc['combination']:
                if len(qc['combination']['Qi']) == 0:
                    qc['combination'].pop('Qi')

        return [max(characteristic_combination, key=lambda x: x['value']),
                max(frequent_combination, key=lambda x: x['value']),
                max(quasi_permanent_combination, key=lambda x: x['value'])]

    def get_result(self, split_result=False):
        """
        EN 1990 (6.14) characteristic (6.15) frequent (6.16) Quasi-permanent

        Combines loads for Service Limit State and returns combination with highest value

        @return: returns three results in a list.
                 First result is for characteristic combination,
                 second is frequent combination,
                 third is quasi permanent combination.
                 Each result consists of two parts: x and y values. Those values are for given ids.
        """
        if len(self.Q_k) + len(self.G_k) <= 0:
            raise Exception('Give at least one Load with add method')
        results = self.combine()
        if split_result:
            final_result = {'char': self.get_single_split_result(results[0]),
                            'freq': self.get_single_split_result((results[1])),
                            'quasi': self.get_single_split_result(results[2])}
        else:
            final_result = {'char': self.get_single_result(results[0]),
                            'freq': self.get_single_result((results[1])),
                            'quasi': self.get_single_result(results[2])}
        return final_result


class ACCCombiner(Combiner):
    def combine(self):
        vals = []
        case1 = []
        for l in self.G_k:
            case1.append({'load': l, 'factor': 1})

        for l in self.Q_k:
            case2q1 = []
            case2qi = []
            if isinstance(l[0], (WindLoad, SnowLoad)):
                case2q1.append({'load': l, 'factor': l[0].comb_factors[1]})
            else:
                case2q1.append({'load': l, 'factor': l[0].comb_factors[2]})

            for lo in self.Q_k:
                if l is not lo:
                    case2qi.append({'load': lo, 'factor': lo[0].comb_factors[2]})
            if isinstance(l[0], (WindLoad, SnowLoad)):
                vals.append({'value': self.sum_G_k() + l[0].combined(1) + self.sum_Q_k_excluding(l[0], 2),
                            'combination': {'G': case1, 'Q1': case2q1, 'Qi': case2qi}})
            else:
                vals.append({'value': self.sum_G_k() + l[0].combined(2) + self.sum_Q_k_excluding(l[0], 2),
                             'combination': {'G': case1, 'Q1': case2q1, 'Qi': case2qi}})
        else:
            vals.append({'value': self.sum_G_k(), 'combination': {'G': case1}})

        for cc in vals:
            if 'G' in cc['combination']:
                if len(cc['combination']['G']) == 0:
                    cc['combination'].pop('G')
            if 'Q1' in cc['combination']:
                if len(cc['combination']['Q1']) == 0:
                    cc['combination'].pop('Q1')
            if 'Qi' in cc['combination']:
                if len(cc['combination']['Qi']) == 0:
                    cc['combination'].pop('Qi')

        return max(vals, key=lambda x: x['value'])

    def get_result(self):
        """

        @return: returns the combination with the highest value
        """
        if len(self.Q_k) + len(self.G_k) <= 0:
            raise Exception('Give at least one Load with add method')
        res = self.combine()
        return self.get_single_result(res)


class MultiCombiner:
    def __init__(self, structure_class='E', CC=2, kk_jako=1):
        self.ULS = ULSCombiner(structure_class, CC, kk_jako)
        self.SLS = SLSCombiner(structure_class, CC, kk_jako)
        self.Acc = ACCCombiner(structure_class, CC, kk_jako)

    def add(self, *args):
        for arg in args:
            self.ULS.add(arg)
            self.SLS.add(arg)
            self.Acc.add(arg)

    def use_reduction_factor(self, val):
        self.ULS.use_reduction_factor(val)
        self.SLS.use_reduction_factor(val)
        self.Acc.use_reduction_factor(val)

    def get_result(self):
        uls_res = self.ULS.get_result()
        sls_res = self.SLS.get_result()
        acc_res = self.Acc.get_result()
        return {'uls': uls_res, 'sls': sls_res, 'acc': acc_res}


if __name__ == "__main__":
    import sections.timber.timber_section as ts

    SW = Load(9.31)

    ont = AreaLoad(2.6)
    vesi = AreaLoad(0.5)
    Qs = SnowLoad(2.0)
    Qw = WindLoad(0.5)
    Hk = ImposedLoad(3)

    M = MultiCombiner('A', kk_jako=8)
    M.add([SW, 'y', [0]])
    M.add([ont, 'y', [0]])
    M.add([Qs, 'y', [0]])
    M.add([vesi, 'y', [0]])
    M.add([Hk, 'y', [0]])
    M.use_reduction_factor(16)
    result = M.get_result()

    # result[ULS/SLS/Acc][x/y][index]
    print(f'ULS result: x: {result["uls"]["x"][0]}, y: {result["uls"]["y"][0]}')

    # result[SLS][ominais/tavallinen/pitkäaikais][x/y][index]
    print(f'SLS result: ominais: x: {result["sls"]["char"]["x"][0]}, y: {result["sls"]["char"]["y"][0]}')
    print(f'SLS result: tavallinen: x: {result["sls"]["freq"]["x"][0]}, y: {result["sls"]["freq"]["y"][0]}')
    print(f'SLS result: pitkäaikais: x: {result["sls"]["quasi"]["x"][0]}, y: {result["sls"]["quasi"]["y"][0]}')
    #print(f'Acc result: {result[2]}')
    print(f'Acc result: x: {result["acc"]["x"][0]}, y: {result["acc"]["y"][0]}')

    # ULS = ULSCombiner('E', kk_jako=8)
    # ULS.add([SW, 'y', [0]])
    # #ULS.add([Hk, 'y', [0]])
    # ULS.add([Qs, 'y', [0]])
    # ULS.add([ont, 'y', [0]])
    # #ULS.add([Qw, 'x', [0, 4]])
    # #ULS.add([ont, 'y', [0]])
    # #ULS.use_reduction_factor(16)
    # res = ULS.get_result()
    # print(res)
    # #
    # SLS = SLSCombiner('A', kk_jako=8)
    # SLS.add([SW, 'y', [0]])
    # #SLS.add([Qw, 'x', [0,4]])
    # #SLS.add([Hk, 'y', [0]])
    # SLS.add([Qs, 'y', [0]])
    # SLS.add([ont, 'y', [0]])
    # #SLS.add([ont, 'y', [0]])
    # #SLS.use_reduction_factor(16)
    # ress = SLS.get_result(True)
    # print(ress['char'])
    #
    # Acc = ACCCombiner('A', kk_jako=10)
    # Acc.add([SW, 'y', [0]])
    # #Acc.add([Qw, 'x', [0]])
    # Acc.add([Qs, 'y', [0]])
    # Acc.add([Hk, 'y', [0]])
    # Acc.add([ont, 'y', [0]])
    # #Acc.add([Hk, 'y', [0]])
    # #Acc.use_reduction_factor(16)
    # resss = Acc.get_result()
    # print(resss)
    # # #print(Acc.combine())




# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Wed Dec 11 19:41:00 2019

@author: kmela
"""


import math
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as lines
import os

from metku.sections.steel.ISection import HEA, HEB, IPE
from metku.structures.steel.plates import RectPlate
from metku.eurocodes.en1993.constants import gammaM0
from metku.eurocodes.en1993.en1993_1_8.en1993_1_8 import Bolt, BoltRow, BoltRowGroup, stiffness_ratio, full_strength_weld_throat
from metku.eurocodes.en1993.en1993_1_8.en1993_1_8 import END_ROW, INNER_ROW, OTHER_INNER_ROW, ROW_OUTSIDE_BEAM_TENSION_FLANGE, FIRST_ROW_BELOW_BEAM_TENSION_FLANGE, OTHER_END_ROW
from metku.eurocodes.en1993.en1993_1_8.en1993_1_8 import END_ROW_ADJACENT_TO_STIFFENER, ADJACENT_TO_STIFFENER_ROW
from metku.eurocodes.en1993.en1993_1_8.en1993_1_8 import TENSION_ROW, SHEAR_ROW, COMBI_ROW
from metku.eurocodes.en1993.en1993_1_8.en1993_1_8 import nut_size, washer_size
import metku.eurocodes.en1993.en1993_1_8.component_method as cm

from metku.cost.workshop import Workshop
from metku.cost.cost_data import BASIC_STEEL_PRICE, steel_grade_add_on, thickness_add_on
from metku.cost.cost_data import bolt_unit_cost_full_thread as bolt_costs_ft
from metku.cost.cost_data import bolt_unit_cost_part_thread as bolt_costs_pt
from metku.cost.cost_data import nut_unit_cost, washer_unit_cost

class EndPlateJoint:
    """ Class for end-plate connections """
    
    
    
    def __init__(self,column,beam,tp,bp,mat_p,etop,ebottom,bolt,y_bolts,e_bolts,
                 bolt_row_pos, groups=[],group_pos=None, row_types=[],workshop=None):
        """ Contructor
            Input:
                column .. ISection or WISection profile
                beam .. ISection or WISection profile
                tp .. thickness of end plate [mm]
                mat_p .. plate material (e.g. "S355")
                etop .. extension of end plate above tension flange [mm]
                ebottom .. extension of end plate below compression flange
                bolt .. Bolt class object
                y_bolts .. array of vertical positions of bolt rows
                           (origin is located in the centroid of the beam)
                e_bolts .. edge distance of bolts (from centroid of bolt hole)
                            measured to the edge of the end plate [mm]
                bolt_row_pos .. list of dicts notifying the position of the row
                                for column flange in bending and end plate in bending
                                components
                                {"flange": flange_pos, "plate": plate_pos}
                groups .. list of bolt row groups
                group_pos . list of bolt row positions in groups
                    group_pos[i] is a list of length of the number of bolt rows in group 'i'.
                    each item in this list is a bolt row position code, e.g. INNER_ROW, see en1993_1_8.py
        """
        
        """ States, whether or not the column continues above the connection """
        self.continuous_column = True
        
        self.col = column
        self.beam = beam
        
        self.MjEd = 0.0
        self.VjEd = 0.0
        self._MjRd = 0.0
        self._VjRd = 0.0
                
        self._beta = 1.0
        
        # calculate plate height
        hp = etop + ebottom + beam.h        

        
        self.end_plate = RectPlate(bp,hp,tp,material=mat_p)
        self._etop = etop
        self.ebottom = ebottom
        self._ebolts = e_bolts      
        self._xbolts = 0.5*bp-e_bolts
        # Make bolt rows
        self.bolt_rows = []
        self.row_groups = []
        
        # Beam flange weld, throat thickness [mm] (to plate)
        self.weld_f = 5
        # Beam web weld, throat thickness [mm] (to plate)
        self.weld_w = 5
        
        # Plate to column weld, throat thickness [mm]
        # Only, if the plate is welded to the column, i.e. no bolts
        self.weld_p = 5
        
        # Distance between centroids of bolts in a row (equal for all rows)
        #self.p = bp-2*e_bolts
        
        # Geometrical properties
        
        # Create bolt rows
        for y, row_pos, r_type in zip(y_bolts,bolt_row_pos,row_types):
            """ Determine location of bolt row with respect to
                column flange in bending and
                end plate in bending components
            """
            flange_pos = row_pos["flange"]
            end_plate_pos = row_pos["plate"]
            
            """
            if y > 0.5*beam.h:
                "" Row is above the beam, so it is an end row
                    for column flange and row outside beam tension
                    flange for end plate
                ""
                flange_pos = END_ROW
                end_plate_pos = ROW_OUTSIDE_BEAM_TENSION_FLANGE
                
            else:
                "" Rows below top (tension) flange of the beam:
                    There are several possibilities:
                        - only one bolt row below the beam flange
                        - more than one rows below the beam flange
                        - 
                ""
            """
            self.bolt_rows.append(BoltRow(bolt,self.p,y,flange_pos,end_plate_pos,joint=self,row_type=r_type))
    
        # Create bolt row groups
        for i, group in enumerate(groups):
            self.row_groups.append(BoltRowGroup([self.bolt_rows[g] for g in group],
                                                group_pos[i])
                                   )
        
        self.stiffeners = {"col_top_flange":None,
                           "col_bottom_flange":None,
                           "col_web_plate":None}
        
        self.workshop = workshop
        
        self._cost = {'plate':0.0,                                         
                     'welding':0.0,
                     'holes':0.0,
                     'bolts':0.0}
    
    @property
    def MjEd(self):        
        return self.beam.Med[0]
    
    @property
    def VjEd(self):
        return self.beam.Ved[1]
    
    @MjEd.setter
    def MjEd(self,val):        
        self.beam.Med[0] = val
    
    @VjEd.setter
    def VjEd(self,val):
        self.beam.Ved[1] = val
    
    @property
    def MjRd(self):
        self._MjRd = self.bending_resistance()
        #if self._MjRd <= 0.0:
        #    MRd = self.bending_resistance()
        return self._MjRd
    
    @property
    def VjRd(self):
        self._VjRd = self.shear_resistance()
        #if self._VjRd <= 0.0:
        #    VRd = self.shear_resistance()
        return self._VjRd
    
    @property
    def bolt_size(self):
        return self.bolt_rows[0].bolt.d
    
    @bolt_size.setter
    def bolt_size(self,val):
        for row in self.bolt_rows:
            row.bolt_size = val
    
    def calculate(self):
        """ This method is for optimization:
            In OptimizationProblem class, constraint evaluation includes
            'fea_required' flag which signifies whether or not finite element
            analysis is performed. The 'fea' method of OptimizationProblem
            class includes the command 'structure.calculate()' which actually
            performs the analysis.
            
            For an end plate joint, the 'calculate' method is to evaluate
            the moment resistance. The idea is that evaluation of initial
            stiffness and shear resistance require knowing the effective lengths,
            etc. which are calculated during evaluating the bending
            resistance.
        """
        
        self.bending_resistance()
        self.shear_resistance()
    
    @property
    def p(self):
        """ Horizontal distance between bolt rows """
        return self.bp-2*self.ebolts
    
    @property
    def ebolts(self):
        return self._ebolts
    
    @ebolts.setter
    def ebolts(self,val):
        """ Distance of bolt center from the edge of the end plate """
        self._ebolts = val
        # Set the distance of bolts in each row based on the
        # new edge distance.
        #for row in self.bolt_rows:
        #    row.p = self.p
    
    @property
    def etop(self):
        return self._etop
    
    @etop.setter
    def etop(self,val):
        self._etop = val
        self.end_plate.h = val + self.ebottom + self.beam.h
    
    @property
    def xbolts(self):
        """ Distance of bolts from center of beam """
        return self._xbolts

    @xbolts.setter
    def xbolts(self,val):
        """ Distance of bolts from center of beam """
        self._xbolts = val
        self.ebolts = 0.5*self.bp-val
    
    @property
    def nrows(self):
        """ Number of bolt rows """
        return len(self.bolt_rows)
    
    @property
    def tension_rows(self):
        """ Number of rows taking tension """
        n = 0
        
        for row in self.bolt_rows:
            if row.type != SHEAR_ROW:
                n += 1
        
        return n
    
    @property
    def zc(self):
        """ Center of compression: located at the center line of the bottom flange
            of the beam
        """
        return -0.5*(self.beam.h-self.beam.tf)
        
    
    @property
    def bp(self):
        """ Width of end plate """
        return self.end_plate.b
    
    @bp.setter
    def bp(self,val):
        """ Set width of end plate """
        self.end_plate.b = val
    
    @property
    def tp(self):
        """ Thickness of end plate """
        return self.end_plate.t
    
    @tp.setter
    def tp(self,val):
        self.end_plate.t = val
    
    
    def emin(self,end_plate_extension=False):
        """ Distance emin used for T-stubs 
            See EN 1993-1-8, 
            6.2.6.4.1(2) for unstiffened column flange
            6.2.6.4.2(4) for stiffened column flange, and
            6.2.6.5(3) for end plates
            
            They all refer to Fig. 6.8. of EN 1993-1-8
            
        """
        
        if end_plate_extension:
            emin = self.ex()
        
        return emin
        
    @property
    def beta(self):
        """ Effect of shear in the column web to moment resistance of the joint 
            So far, only one value is given (approximation for one-sided joint)
        """
        return self._beta
    
    @beta.setter
    def beta(self,val):
        self._beta = val
    
    def Mpl_fc_Rd(self):
        """ Plastic moment resistance of column flange 
            This is needed, if transverse stiffeners are used on both
            the tension and compression sides
        """
        
        return 0.25*self.col.b*self.col.tf**2*self.col.fy/gammaM0
        
    def Mpl_st_Rd(self,ts,fy):
        """ Plastic moment resistance of a stiffener
            This is needed, if transverse stiffeners are used on both
            the tension and compression sides
        """
        
        return 0.25*self.col.b*ts**2*fy/gammaM0
    
    def add_stiffener(self,stiffener,**kwargs):
        """ Adds a stiffener plate to the joint """
        
        for key, value in kwargs.items():
            if key == 't':
                if stiffener.find('flange') != -1:
                    hp = self.col.h-2*self.col.tf
                    bp = 0.5*(self.col.b-self.col.tw)                    
                else:
                    hp = self.beam.h
                    bp = self.col.hw
                tp = value
                plate = RectPlate(hp,bp,tp,material=self.col.material.name)
            elif key == 'plate':
                plate = value
            
            plate.weld_size = math.ceil(full_strength_weld_throat(plate.material)*plate.t)
            
            self.stiffeners[stiffener] = plate
            
            if stiffener == 'col_top_flange':
                """ Modify joint as follows:
                    1. Find the top row. If it is above the stiffener,
                        delete all groups that contain that row.
                        The role of the row in column flange in bending
                        becomes End bolt row adjacent to a stiffener
                    2. The role of of the first row below the stiffener
                        becomes 'Bolt-row adjacent to a stiffener'
                """
                
                """ It is assumed that the bolt rows are ordered in descending
                    value of z coordinate, i.e. row[0] is the top row.
                """
                
                if self.bolt_rows[0].z > 0.5*self.beam.h:
                    """ In this case the first row is above the top flange of the beam """
                    if self.continuous_column:                
                        self.bolt_rows[0].loc_col_flange = ADJACENT_TO_STIFFENER_ROW
                    else:
                        self.bolt_rows[0].loc_col_flange = END_ROW_ADJACENT_TO_STIFFENER
                    self.bolt_rows[0].Tstub_col_flange.m2 = self.bolt_rows[0].top_flange_stiffener_m2
                    self.bolt_rows[1].loc_col_flange = ADJACENT_TO_STIFFENER_ROW
                    self.bolt_rows[1].Tstub_col_flange.m2 = self.bolt_rows[1].top_flange_stiffener_m2
                    
                    for group in self.row_groups:
                        for i, row in enumerate(group.rows):
                            if row == self.bolt_rows[0]:   
                                if self.continuous_column:
                                    group.row_loc[i]['flange'] = ADJACENT_TO_STIFFENER_ROW
                                else:
                                    group.row_loc[i]['flange'] = END_ROW_ADJACENT_TO_STIFFENER
                            elif row == self.bolt_rows[1]: 
                                
                                group.row_loc[i]['flange'] = ADJACENT_TO_STIFFENER_ROW
                                
                        
                else:
                    """ In this case the first row is below the top flange of the beam """
                    self.bolt_rows[0].loc_col_flange = ADJACENT_TO_STIFFENER_ROW
                pass
                
                
        
    def V_wp_Rd(self):
        """ Column web shear resistance """
        V_wp_add_Rd = [0,0]
        if self.stiffeners['col_top_flange'] is not None and self.stiffeners['col_bottom_flange'] is not None:
            """ Distance between center lines of the stiffeners:
                it is assumed that the stiffeners are located at
                beam flanges (this may not be true for top stiffener for flushed end-plat)
            """
            ds = self.beam.h-self.beam.tf            
            V_wp_add_Rd[0] = 4*self.Mpl_fc_Rd()
            fyt = self.stiffeners['col_top_flange'].fy
            tst = self.stiffeners['col_top_flange'].t
            fyb = self.stiffeners['col_bottom_flange'].fy
            tsb = self.stiffeners['col_bottom_flange'].t
            V_wp_add_Rd[1] = (2*self.Mpl_fc_Rd() + self.Mpl_st_Rd(tst,fyt) + self.Mpl_st_Rd(tsb,fyb))/ds
            
            
        return cm.col_web_shear(self.col) + min(V_wp_add_Rd)
    
    def k1(self,z=None):
        """ Stiffnes factor of column web in shear """
        if z is None:
            z = self.zeq()
    
        return cm.col_web_shear_k(self.col,self.beta,z)
            
    
    def Fc_wc_Rd(self,verb=False):
        """ Column web in compression """
        
        if self.stiffeners["col_bottom_flange"] is None:
            # Calculate compression stress
            sigma_com_Ed = self.col.sigma_com()
            Fc_wc_Rd, beff_wc = cm.col_web_trv_comp(self.col, self.beam, self.end_plate.t, self.ebottom, self.weld_f, self.beta, sigma_com_Ed, verb)
            self.beff_wc = beff_wc
        else:
            Fc_wc_Rd = cm.col_web_trv_comp_stiffened(self.col, self.stiffeners["col_bottom_flange"])
        #print("beff_wc_Rd = {0:4.2f} mm".format(beff_wc))
        return Fc_wc_Rd
    
    def beff_c_wc(self):
        
        # Column information
        r_c = self.col.r                                     # Column rounding, VALSSATTU
        t_fc = self.col.tf                                  # Column flange thickness
        t_wc = self.col.tw                                  # Column web thickness
        d_wc = self.col.hw                                  # Column web height
        #A_vc = self.col.Ashear                                  # Shear area of column web
        #f_y_wc = self.col.fy                                # Yield strength of column material
        #E = self.col.column.E                                  # Young's modulus of column material

        # Beam information
        t_fb = self.beam.tf                           # Beam flange thickness

        # Plate information
        t_p = self.tp
        lo_p = self.ebottom
        
        # weld between plate and bottom flange
        a_p = self.weld_f            

        # Width of compressed zone under compressed flange using 45 deg stress distribution
        s_p = min(2.0*t_p, t_p+lo_p-math.sqrt(2.0)*a_p)           # lo_p is overhang of end-plate over lower flange of beam

        # Effective length of compressed web
        s = r_c
        b_eff_c_wc = t_fb+2.0*math.sqrt(2.0)*a_p+5.0*(t_fc+s)+s_p
        
        return b_eff_c_wc
    
    
    def k2(self):
        """ Stiffness factor for column web in compression """
        
        if self.stiffeners["col_bottom_flange"] is None:
            k = cm.col_web_trv_comp_k(self.col,self.beff_c_wc())
        else:
            k = np.inf
        return k
    
    def Fc_fb_Rd(self):
        """ Beam flange and web in compression """        
        return cm.beam_web_compression(self.beam)
    
    def bending_resistance(self,verb=False):
        """ Bending moment resistance """
        
        # Calculate column web shear resistance
        # and compression resistances
        V_wp_Rd = self.V_wp_Rd()
        Fc_wc_Rd = self.Fc_wc_Rd()
        Fc_fb_Rd = self.Fc_fb_Rd()
        
        Fcom_Rd = min(V_wp_Rd/self.beta,Fc_wc_Rd,Fc_fb_Rd)
        
        if verb:
            print("Compression side:")
            print("Column web in shear: V_wp_Rd = {0:4.2f} kN".format(V_wp_Rd*1e-3))
            print("Transformation parameter: beta = {0:4.3f} kN".format(self.beta))
            print("Column web in shear: V_wp_Rd/beta = {0:4.2f} kN".format(V_wp_Rd*1e-3/self.beta))
            print("Column web in compression: Fc_wc_Rd = {0:4.2f} kN".format(Fc_wc_Rd*1e-3))
            print("Beam web in compression: Fc_fb_Rd = {0:4.2f} kN".format(Fc_fb_Rd*1e-3))
            print("Compression resistance: Fcom_Rd = {0:4.2f} kN".format(Fcom_Rd*1e-3))
        # Sum of tension forces in the bolt rows
        Ft_total = 0.0
        
        # Calculate resistances of bolt rows
        for i, row in enumerate(self.bolt_rows):
            if row.type != SHEAR_ROW:
                if verb:
                    print("  ** ROW {0:1.0f} **".format(i+1))
        
                FtRd = row.tension_resistance(verb)
                
                """ The total force in the bolt rows, including
                    the current bolt row, cannot exceed the
                    compression resistance (including shear of column web panel)
                    of the joint
                """
                #row.FtRd = min(FtRd,Fcom_Rd-Ft_total)
                
                #Ft_total += row.FtRd                
        
        if verb:
            print("--------  ROW RESISTANCES ---------")
            for i, row in enumerate(self.bolt_rows):            
                if row.type != SHEAR_ROW:
                    print("ROW {0:1.0f}: FtRd = {1:4.2f} kN".format(i+1, row.FtRd*1e-3))
            print("-----------------------------------")
        # Calculate resistances of bolt row groups 
        if verb:
            print("*** BOLT ROW GROUPS ***")
        
        for i, group in enumerate(self.row_groups):
            if verb:
                print("  * GROUP {0:4.0f} * ".format(i+1))
    
            Ft_group = group.tension_resistance(verb)
            
            if verb:
                print("Ft_Group = {0:4.2f} kN".format(Ft_group*1e-3))
                #print("Ft_rows = {0:4.2f} kN".format(Ft_tot*1e-3))
            
            # Sum of forces of the bolt rows of the group, considered
            # individually
                        
            #Ft_tot = sum([row.FtRd for row in group.rows])
            Ft_tot = 0.0
            
            for row in group.rows:
                row.FtRd = min(row.FtRd,Ft_group-Ft_tot)
                Ft_tot += row.FtRd
            
            """
            Fred = Ft_tot - Ft_group
            if Fred > 0:
                "" In this case, the sum of resistances of the bolt rows
                    in the group is greater than the group resistance:
                        the forces in the group must be reduced
                ""
                for row in group.rows:
                    dF = row.FtRd-Fred
                    if dF >= 0:
                        "" Fred can be compensated by reducing the strength
                            of the current row.
                        ""
                        row.FtRd = dF
                        break
                    else:
                        "" The capacity of the current row is not enough to
                            cover the difference between sum of bolt row
                            resistances and resistance of the group
                        ""
                        
                        "" Subtract the resistance of the current row from
                            the total force to be compensated. The remaining
                            force, if any, needs to be compensated by further rows
                        ""
                        Fred -= row.FtRd
                        row.FtRd = 0.0
            """
            #print("Need to reduce strength of rows")
        
        if verb:
            print("--------  ROW RESISTANCES AFTER GROUPS ---------")
            for i, row in enumerate(self.bolt_rows):            
                if row.type != SHEAR_ROW:
                    print("ROW {0:1.0f}: FtRd = {1:4.2f} kN".format(i+1, row.FtRd*1e-3))
                
        Ft_total = 0.0
        for row in self.bolt_rows:
            if row.type != SHEAR_ROW:                 
                row.FtRd = min(row.FtRd,Fcom_Rd-Ft_total)
                Ft_total += row.FtRd
        
        if verb:
            print("--------  ROW RESISTANCES COMPRESSION ---------")
            for i, row in enumerate(self.bolt_rows):            
                if row.type != SHEAR_ROW:
                    print("ROW {0:1.0f}: FtRd = {1:4.2f} kN".format(i+1, row.FtRd*1e-3))
        
        MjRd = 0.0
        for row in self.bolt_rows:
            MjRd += row.FtRd * row.h
        
        if verb:
            print("MjRd = {0:4.2f} kNm".format(MjRd*1e-6))
        
        self._MjRd = MjRd
        
        return MjRd
    
    def hr(self,which_rows='all'):
        """ Returns an array of bolt row distances from 
            Compression centroid
        """
        if which_rows == 'all':
            hr = np.array([row.h for row in self.bolt_rows])
        elif which_rows == 'non-shear':
            hr = np.array([row.h for row in self.bolt_rows if row.type != SHEAR_ROW])
        
        return hr

    def keq(self,verb=False):
        """ Equivalent stiffness of the tension rows """
        keff = np.array([row.keff() for row in self.bolt_rows])
        hr = self.hr()
        
    
        return sum(keff*hr)/self.zeq(verb)
    
    def zeq(self,verb=False):
        """ Equivalent moment arm """
        zeq = 0.0
        
        keff = np.array([row.keff() for row in self.bolt_rows if row.type != SHEAR_ROW])
        hr = self.hr('non-shear')
        
        zeq = sum(keff*hr**2)/sum(keff*hr)
        
        return zeq
    
    def zeq_and_keq(self,verb=False):
        """ Return both zeq and keq 
            To avoid double computations
        """
        zeq = 0.0
        
        keff = np.array([row.keff() for row in self.bolt_rows if row.type != SHEAR_ROW])
        hr = self.hr('non-shear')
        
        zeq = sum(keff*hr**2)/sum(keff*hr)
        keq = sum(keff*hr)/zeq
        
        if verb:
            print(" * zeq and keq *")
            for k, h in zip(keff,hr):
                print("keff = {0:4.2f} mm; hr = {1:4.2f} mm".format(k,h))
            
        
        return zeq, keq
        
    def moment_arm(self):
        """ Moment arm for joints with two bolt rows or """
        
        """ hr is an array of distances of bolt rows from
            compression center
        """
        hr = self.hr()
        
        return max(hr)
            

    def k_comp(self,z):
        """ Combined stiffness of compression side components,
            namely
            column web in shear and
            column web in compression
        """
        k1 = self.k1(z)
        k2 = self.k2()
        
        return 1/(1/k1+1/k2)
        

    def Sj_ini(self,verb=False):
        """ Initial rotational stiffness """
        
        if verb:
            print("*** INITIAL ROTATIONAL STIFFNESS ***")
            
        
        E = self.col.material.E
        
        if self.tension_rows == 1:
            """ Only one tension row: moment arm is the distance
                between bottom 
            """
            z = self.moment_arm()
        else:
            z, k_eq = self.zeq_and_keq(verb)
        
        kcomp = self.k_comp(z)
        
        if verb:
            print("   Moment arm: z = {0:4.2f} mm".format(z))
            print("   Stiffness factors:")
            print("     k1 = {0:4.2f} mm".format(self.k1(z)))
            print("     k2 = {0:4.2f} mm".format(self.k2()))
            print("     keq = {0:4.2f} mm".format(k_eq))
            print("     kcomp = {0:4.2f} mm".format(kcomp))
        
        Sj_ini = E*z**2/(1/kcomp+1/k_eq)
        
        if verb:
            print("  Sj_ini = {0:4.2f} kNm/rad".format(Sj_ini*1e-6))
        
        return Sj_ini 
    
    def rotational_stiffness(self,verb=False):
        """ Calculated rotational stiffness """
        
        MjRatio = self.MjEd/(self.MjRd*1e-6)
        
        mu = stiffness_ratio(MjRatio,psi=2.7)
        
        Sj = self.Sj_ini()/mu
        
        if verb:
            print("--- Rotational stiffness ---")
            print(" Initial: Sj_ini = {0:4.2f} kNm/rad".format(self.Sj_ini()*1e-6))
            print(" Moment ratio: MjEd/MjRd = {0:4.2f}".format(MjRatio))
            print(" Stiffness ratio: mu = {0:4.2f}".format(mu))
            print(" Stiffness: Sj = {0:4.2f}".format(Sj*1e-6))
        
        return Sj
    
    def shear_resistance(self,verb=False):
        """ Shear resistance of the joint """
        
        VRd = 0.0
        
        if verb:
            print("-- SHEAR RESISTANCE --")
        
        for i, row in enumerate(self.bolt_rows):
            if row.FtRd > 0.0:
                """ Rows in tension still have at least 28.6% of their
                    shear resistance left.
                """
                
                # Bearing resistance of end plate
                e = [0,self.ebolts]
                p = [0,self.p]
                if row.loc_end_plate == ROW_OUTSIDE_BEAM_TENSION_FLANGE:
                    e[0] = self.etop
                    lpos = 'edge'
                else:
                    lpos = 'inner'
                    p[0] = abs(self.bolt_rows[i-1].z - row.z)
                    
                VbRd_plate = row.bolt.bearing_resistance(self.end_plate.fu,
                                                         self.tp,
                                                         e,p,pos_perp="edge",pos_load=lpos,verb=False)
                # Bearing resistance of column flange
                if row.loc_end_plate == ROW_OUTSIDE_BEAM_TENSION_FLANGE:
                    e[0] = 1e5
                    lpos = 'edge'
                else:
                    lpos = 'inner'
                    p[0] = abs(self.bolt_rows[i-1].z - row.z)
                    
                VbRd_col = row.bolt.bearing_resistance(self.col.fu,
                                                         self.tp,
                                                         e,p,pos_perp="edge",pos_load=lpos,verb=False)
                
                # Bearing resistance
                VbRd = min(VbRd_plate,VbRd_col)
                
                
                VRd += 2*0.286*row.bolt.shear_resistance()
            elif abs(row.FtRd) < 1e-3 or row.type == SHEAR_ROW:
                VRd += 2*row.bolt.shear_resistance()
        
        if verb:
            print("VjRd = {0:4.2f} kN".format(VRd*1e-3))
        
        self._VjRd = VRd
        
        return VRd
        
        
    def cost(self,workshop=None,material_cost=BASIC_STEEL_PRICE,verb=False):
        """ Calculate cost of the joint
            units: e
        """
        
        for key, value in self._cost.items():
            self._cost[key] = 0.0
        
        if workshop is not None:
            self.workshop = workshop        
        
        plate_cost = self.workshop.steel_price["plates"]
        
        self._cost['plate'] = self.end_plate.cost(self.workshop,plate_cost)
        hole_cut_length = 0.0
        for row in self.bolt_rows:
            """ Costs for bolts and hole forming 
                bolt costs inlucde one nut and two washers
            """            
            if row.bolt.bolt_type == "Full thread":
                bolt_cost = 2*bolt_costs_ft[int(row.bolt.length)][int(row.bolt.d)]
            else:
                bolt_cost = 2*bolt_costs_pt[int(row.bolt.length)][int(row.bolt.d)]
            self._cost['bolts'] += bolt_cost
            self._cost['bolts'] += nut_unit_cost[int(row.bolt.d)]
            self._cost['bolts'] += washer_unit_cost[int(row.bolt.d)]
            """ Hole costs include forming holes both in the plate in the column """
            hole_cut_length += 2*row.bolt.d0*math.pi
           
            #self._cost['holes'] += 2*workshop.cost_centres['cutting'].cost(self.col.tf,row.bolt.d0*math.pi)

        self._cost['holes'] = self.workshop.cost_centres['cutting'].cost(self.tp,hole_cut_length)

        """ Welding cost """
        flange_weld_length = 2*(self.beam.b + self.beam.tf) - 2*self.beam.r - self.beam.tw + self.beam.r*math.pi
        #print(flange_weld_length)
        self._cost['welding'] += self.workshop.cost_centres['assembly_welding'].cost(self.weld_f,2*flange_weld_length)
        
        web_weld_length = 2*self.beam.hw
        #print(web_weld_length)
        self._cost['welding'] += self.workshop.cost_centres['assembly_welding'].cost(self.weld_w,web_weld_length)
        
        # Add cost for stiffeners
        for stiffener, plate in self.stiffeners.items():
            if plate is not None:
                self._cost[stiffener] = plate.cost(self.workshop,plate_cost)
            
        total_cost = 0.0
        for c in self._cost.values():
            total_cost += c
        
        self._total_cost = total_cost
        
        #self.cost_distribution()
        
        if verb:
            self.cost_distribution()
        
        return total_cost
    
    def cost_distribution(self,pie=False):
        """ Print cost distribution """
        
        if pie:
            pass
        else:
            print(""" Cost distribution of the connection """)
            p_tot = 0.0
            for key, value in self._cost.items():
                p =  value/self._total_cost*100
                p_tot += p
                print("   {0:10s}: {1:5.2f} € [{2:5.2f} %]".format(key,value,p))
            
            print("-----------------------")
            print("   {0:10s}: {1:5.2f} € [{2:5.2f} %]".format("Total",self._total_cost,p_tot))
    
    def info(self,draw=False):
        """ Print relevant information of the joint """
        
        print("Resistance of connection:")
        print(" V_wp_Rd = {0:{fm}} [kN] (Column web panel in shear)".format(self.V_wp_Rd()*1.0e-3, fm='7.2f'))
        print("F_c_wc_Rd = {0:{fm}} [kN] (Column web in transverse compression)".format(self.Fc_wc_Rd()*1.0e-3, fm='7.2f'))
        print("F_c_fb_Rd = {0:{fm}} [kN] (Beam flange and web in compression)".format(self.Fc_fb_Rd()*1.0e-3, fm='7.2f'))
        
        if draw:
            self.draw()
        
        
    def draw(self, name=None, two_views=False):
        """ Draw the connection 
            ax[1] .. front view
            ax[0] .. side view
        """
        
        grey = '0.8'
        center_line_col = '0.3'
        if two_views:
            fig, ax = plt.subplots(1,2)
            ax_side = ax[0]
            ax_front = ax[1]
        else:
            fig, ax = plt.subplots(1,1)
            ax_side = ax
            ax_front = ax
        ymax = 150
                        
        t_fc = self.col.tf
        r_c = self.col.r
        h_c = self.col.h
        b_c = self.col.b
        h_b = self.beam.h
        b_b = self.beam.b
        t_fb = self.beam.tf
        t_wb = self.beam.tw
        #d_wb = self.beam.d_w
        a_f = self.weld_f
        r_b = self.beam.r
        h_p = self.end_plate.h
        t_p = self.end_plate.t
        b_p = self.end_plate.b
        up_p = self.etop
        bot_p = self.ebottom
        
        #self.column.draw(axes=ax[1])
        
        # Side view:        
        # Origin 
        # Draw column
        #plt.figure("side " + str(self.plot_nro))
        #plt.clf()
        #plt.xlabel("[mm]")
        #plt.ylabel("[mm]")

        axis_lim = max(h_c,h_b,h_p) + 200
        #axis_lim = max(h_c + 500.0, h_b + 500.0, h_p + 200.0)

        # Plot column in figure
        column_y = [-100.0, axis_lim - 100.0]
        
        ax_side.vlines([0,-t_fc,-(h_c-t_fc),-h_c],column_y[0],column_y[1],'k',zorder=2)
        ax_side.vlines([-(t_fc + r_c),-(h_c - (t_fc + r_c))],column_y[0],column_y[1],grey,zorder=1)
        # Center line
        ax_side.vlines([-0.5*h_c],column_y[0],column_y[1],center_line_col,linestyle='dashdot',zorder=1)
        
        
        # Plot plate in figure
        plate_x = [0.0, t_p, t_p, 0.0]
        plate_y = [0.0, 0.0, h_p, h_p]
        ax_side.plot(plate_x, plate_y, 'k', zorder=2)
        
        # Plot beam in figure
        beam_lim = axis_lim - (100.0 + h_c + t_p)
        beam_x = [t_p, beam_lim]
        ax_side.hlines([bot_p,bot_p+t_fb,bot_p+h_b,bot_p+h_b-t_fb],beam_x[0],beam_x[1],'k',zorder=2)
        ax_side.hlines([bot_p+t_fb+r_b,bot_p+h_b-t_fb-r_b],beam_x[0],beam_x[1],grey,zorder=1)
        ax_side.hlines([bot_p+0.5*h_b],-0.5*h_c,beam_x[1],center_line_col,linestyle='dashdot',zorder=1)
        # Plot welds
        ax_side.plot([t_p, t_p + math.sqrt(2.0)*a_f], [bot_p + h_b + math.sqrt(2.0)*a_f, bot_p+h_b], 'k', zorder=1)
        ax_side.plot([t_p, t_p + math.sqrt(2.0)*a_f], [bot_p + h_b - t_fb - math.sqrt(2.0)*a_f, bot_p + h_b -t_fb], 'k', zorder=1)
        ax_side.plot([t_p, t_p + math.sqrt(2.0)*a_f], [bot_p - math.sqrt(2.0)*a_f, bot_p], 'k', zorder=1)
        ax_side.plot([t_p, t_p + math.sqrt(2.0)*a_f], [bot_p + t_fb + math.sqrt(2.0)*a_f, bot_p + t_fb], 'k', zorder=1)
        
        z0 = self.ebottom + 0.5*h_b
        
        for row in self.bolt_rows:
            h_bolt = row.bolt.head_t
            d_m = row.bolt.head_d
            t_washer = row.bolt.washer_t
            h_nut = row.bolt.nut_t
            L_b = row.bolt.length
            z = row.z + z0
            d = row.bolt.d0
            # Plot bolt hole
            ax_side.plot([-t_fc, t_p], [z + 0.5*d, z + 0.5*d], 'k--', linewidth=0.6)
            ax_side.plot([-t_fc, t_p], [z - 0.5*d, z - 0.5*d], 'k--', linewidth=0.6)
        
            # Plot bolt head
            bolt_head = patches.Rectangle((t_p, z - 0.5*d_m), width=h_bolt, height=d_m,edgecolor='k', facecolor="w", zorder=2)
            ax_side.add_patch(bolt_head)
            
            nut_head = patches.Rectangle((-t_fc - t_washer - h_nut, z - 0.5*d_m), width=h_nut, height=d_m, edgecolor='k', facecolor="w", zorder=2)
            ax_side.add_patch(nut_head)
        
            # Plot center line
            ax_side.plot([-0.5 * L_b, 0.5 * L_b], [z, z], 'b-.', linewidth=0.6, zorder=3)
        
        """ Plot stiffeners (if any) """
        if self.stiffeners['col_top_flange'] is not None:
            t_stiff = self.stiffeners['col_top_flange'].t
            ax_side.hlines([self.ebottom + h_b-0.5*t_fb-0.5*t_stiff,self.ebottom + h_b-0.5*t_fb+0.5*t_stiff],-h_c+t_fc,-t_fc,'b',zorder=2)
        
        if self.stiffeners['col_bottom_flange'] is not None:
            t_stiff = self.stiffeners['col_bottom_flange'].t
            ax_side.hlines([self.ebottom+0.5*t_fb-0.5*t_stiff,self.ebottom+0.5*t_fb+0.5*t_stiff],-h_c+t_fc,-t_fc,'b',zorder=2)
        
        ax_side.set_aspect('equal', 'datalim')
        
        
        ax_side.set_axis_off()
        
        """ Draw front view """        
        if two_views:
            x_offset = 0.0
        else:
            x_offset = beam_lim + 200        
        
        # Draw column
        col_front_x = [-0.5*b_c, -0.5*b_c,0.5*b_c, 0.5*b_c]
        col_front_x = [x+x_offset for x in col_front_x]
        ax_front.vlines(col_front_x, column_y[0], column_y[1], 'k', linewidth=1.0)

        # Draw plate
        plate_front = patches.Rectangle((-0.5*b_p+x_offset, 0.0), width=b_p, height=h_p,\
                                        linewidth=1.0, edgecolor='k', facecolor="w", zorder=2)
        ax_front.add_patch(plate_front)
        
        # Draw beam profile
        self.beam.draw(axes=ax_front,origin=[x_offset,self.ebottom+0.5*h_b])
        
        # Draw bolt rows
        for row in self.bolt_rows:
            #w = 0.5*row.p
            w = 0.5*row.w
            
            h_bolt = row.bolt.head_t
            d_m = row.bolt.head_d
            d_washer = row.bolt.washer_d
            h_nut = row.bolt.nut_t
            L_b = row.bolt.length
            z = row.z + z0
            d = row.bolt.d0
            
            # Plot washers            
            ax_front.add_patch(patches.Circle((-w+x_offset, z), 0.5*d_washer, edgecolor='k', facecolor="w", zorder=4))
            ax_front.add_patch(patches.Circle((w+x_offset, z), 0.5*d_washer, edgecolor='k', facecolor="w", zorder=4))
            # Plot bolts            
            ax_front.add_patch(patches.Circle((-w+x_offset, z), 0.5*d_m, edgecolor='k', facecolor="w", zorder=4))
            ax_front.add_patch(patches.Circle((w+x_offset, z), 0.5*d_m, edgecolor='k', facecolor="w", zorder=4))
            ax_front.add_patch(patches.Circle((-w+x_offset, z), 0.5*d, linestyle='--', linewidth=0.6, edgecolor='k',
                                                    facecolor="w", zorder=4))
            ax_front.add_patch(patches.Circle((w+x_offset, z), 0.5*d, linestyle='--', linewidth=0.6, edgecolor='k',
                                                    facecolor="w", zorder=4))
            
            ax_front.vlines([-w+x_offset,w+x_offset],z - 0.6 * d_m, z + 0.6 * d_m,'b', linewidth=0.6, zorder=5)
            ax_front.hlines(z,-w - 0.6 * d_m+x_offset, -w + 0.6 * d_m+x_offset,'b', linewidth=0.6, zorder=5)
            ax_front.hlines(z,w - 0.6 * d_m+x_offset, w + 0.6 * d_m+x_offset,'b', linewidth=0.6, zorder=5)
        
            #ax_front.plot([-w, -w], [z - 0.6 * d_m, z + 0.6 * d_m], 'b-.', linewidth=0.6, zorder=5)
            #ax_front.plot([-0.5 * w - 0.6 * d_m, -0.5 * w + 0.6 * d_m], [z, z], 'b-.', linewidth=0.6, zorder=5)
            #ax_front.plot([0.5 * w, 0.5 * w], [z - 0.6 * d_m, z + 0.6 * d_m], 'b-.', linewidth=0.6, zorder=5)
            #ax_front.plot([0.5 * w - 0.6 * d_m, 0.5 * w + 0.6 * d_m], [z, z], 'b-.', linewidth=0.6, zorder=5)
        
        ax_front.set_aspect('equal', 'datalim')
        
        ax_front.set_axis_off()
        
        if name is not None:
            try:
                fig.savefig(name,format='svg',bbox_inches='tight')
            except:
                print("Draw unsuccessful.")
        
        
        
def example_1():
    bolt = Bolt(24,10.9)
    bolt.washer_t = 0.0
    col = HEA(340,fy=235)
    beam = IPE(500,fy=235)
    
    hp = 50+85+80+320+65
    bp = 240
    ebottom = hp-85-beam.h
    
    y0 = ebottom + 0.5*beam.h
    y_bolts = []
    y_bolts.append(hp-50-y0)
    y_bolts.append(hp-50-85-y0)
    y_bolts.append(hp-50-85-80-y0)
    y_bolts.append(hp-50-85-80-320-y0)
    
    bolt_row_pos = [{"flange": END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                    {"flange": INNER_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                    {"flange": END_ROW, "plate": OTHER_END_ROW},
                    {"flange": END_ROW, "plate": OTHER_END_ROW},
                    ]
    bolt_row_groups = [[0,1],[0,1,2],[1,2]]
    
    row_types = [TENSION_ROW,TENSION_ROW,TENSION_ROW,SHEAR_ROW]
    
    # POSITION OF ROWS IN GROUPS
    # For each row, position with respect to column flange and end plate are
    # required.
    group_pos = [[{"flange":END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                  {"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE}
                 ],
                 [{"flange":END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                  {"flange":INNER_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                  {"flange":END_ROW, "plate": OTHER_END_ROW}
                 ],
                 [{"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                  {"flange":END_ROW, "plate": OTHER_END_ROW}
                 ]
                 ]
    conn = EndPlateJoint(col,beam,tp=15,bp=bp,mat_p="S235",etop=85,\
                         ebottom=ebottom,bolt=bolt,y_bolts=y_bolts,\
                         e_bolts=60,\
                         bolt_row_pos=bolt_row_pos,
                         groups=bolt_row_groups,
                         group_pos=group_pos,
                         row_types=row_types)
        
    conn.weld_f = 8
    
    conn.workshop = Workshop()
    
    
    return conn

def example_diaz(MjEd=0.0,VjEd = 0.0,connection='A', rows=2):
    d = 12
    bolt = Bolt(d,8.8)
    col = HEB(140,fy=355)
    
    # Connection A
    if connection == 'A':
        beam = IPE(200,fy=355)
        tp = 7.15
        e = 28.8
        ex = 37.61
        px = 58.01
        MjEd = 22
        VjEd = 26.325
    else:    
        beam = IPE(270,fy=355)
        tp = 11.5
        e = 39.3
        ex = 35.03
        px = 58.69
        MjEd = 40
        VjEd = 32
    
    aw = 4.0
    af = 6.0
        
    bp = col.b    
    ebottom = math.floor(math.sqrt(2)*af + tp)
    
    
    hp = beam.h + 5*d + ebottom
    etop = 5*d
    
    """ Generate bolt rows
        y is the vertical position of a bolt row. The
        origin is located in the centroid of the beam
    """
    y0 = ebottom + 0.5*beam.h
    y_bolts = []
    #y_bolts.append(hp-ex-y0)
    #y_bolts.append(hp-ex-px-y0)
    y_bolts.append(0.5*beam.h + etop-ex)
    y_bolts.append(y_bolts[0]-px)
    
    p_bolts = [50,110,110,170][:rows]
    
    bolt_row_pos = [{"flange": END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                    {"flange": INNER_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                    ]
    bolt_row_groups = [[0,1]]
    
    row_types = [TENSION_ROW,TENSION_ROW]
    
    # POSITION OF ROWS IN GROUPS
    # For each row, position with respect to column flange and end plate are
    # required.
    group_pos = [[{"flange":END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                  {"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE}
                 ]]
    conn = EndPlateJoint(col,beam,tp=tp,bp=bp,mat_p="S355",etop=4*d,\
                         ebottom=ebottom,bolt=bolt,y_bolts=y_bolts,\
                         e_bolts=e,\
                         bolt_row_pos=bolt_row_pos,
                         groups=bolt_row_groups,
                         group_pos=group_pos,
                         row_types=row_types)
    
    conn.weld_f = af
    conn.weld_w = aw
    
    conn.workshop = Workshop()
    
    conn.MjEd = MjEd
    conn.VjEd = VjEd
    
    return conn

def diax_ex_rows(MjEd=40.0,VjEd = 32.0, rows = 4, extended=True):

    steel_grade = 275
    d = 16   
    col = HEB(140,steel_grade)
    beam = IPE(270,fy=steel_grade)
    
    ex = 27
    px = 61
    
    """ Weld size 
        aw .. web
        af .. flange
    """
    aw = 4.0
    af = 6.0
    
    """ Overhang distance
        ebottom .. plate below bottom flange of the beam
        etop .. plate above top flane of the beam
    """
    ebottom = math.ceil(math.sqrt(2)*af) + 2
    if extended:
        etop = ex + 0.5*(px-beam.tf)
    else:
        etop = ebottom
    
    """ End plate dimensions
        hp .. height
        bp .. width
        tp .. thickness
    """
    plate_mat = "S" + str(steel_grade)
    hp = beam.h + etop + ebottom
    bp = 135
    tp = 13.0
    
    
    """ Distance of bolts from the edge of the end plate """    
    e = 32.0   
    
    """ Distances of bolt rows:
        [0] .. distance from top edge of plate
        [i] .. distance between adjacent rows
    """
    if extended:
        p_bolts = [ex,px,px,120][:rows]
    else:
        p_bolts = [etop+beam.tf+1.5*af+0.5*d + 20,80,80,120][:rows]


    """ Generate bolt rows
        y is the vertical position of a bolt row. The
        origin is located in the centroid of the beam
    """
    y0 = ebottom + 0.5*beam.h
    y_bolts = []
    y = 0.5*beam.h + etop
    for p in p_bolts:
        y -= p
        y_bolts.append(y)    
    
    """ Generate bolt row position labels """
    if extended:
        bolt_row_pos = [{"flange": END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                        {"flange": INNER_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                        {"flange": INNER_ROW, "plate": OTHER_END_ROW},
                        {"flange": END_ROW, "plate": OTHER_END_ROW},
                        ]
    else:
        bolt_row_pos = [{"flange": END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                        {"flange": INNER_ROW, "plate": OTHER_INNER_ROW},
                        {"flange": INNER_ROW, "plate": OTHER_END_ROW},
                        {"flange": END_ROW, "plate": OTHER_END_ROW},
                        ]
    bolt_row_pos = bolt_row_pos[:rows]
    bolt_row_pos[-1]["flange"] = END_ROW
    
    """ Generate bolt row groups """
    bolt_row_groups = [[0,1], [0,1,2], [1,2]]
    
    if rows == 2:
        bolt_row_groups = [[0,1]]
    
    row_types = [TENSION_ROW,TENSION_ROW,TENSION_ROW,SHEAR_ROW][:rows]
    
    
    # POSITION OF ROWS IN GROUPS
    # For each row, position with respect to column flange and end plate are
    # required.
    if extended:
        group_pos = [[{"flange":END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                      {"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE}
                     ],
                     [{"flange":END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                      {"flange":INNER_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ],
                     [{"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ]
                     ]
    else:
        group_pos = [[{"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ],
                     [{"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                      {"flange":INNER_ROW, "plate": OTHER_INNER_ROW},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ],
                     [{"flange":END_ROW, "plate": OTHER_END_ROW},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ]
                     ]
        
    if rows == 2:
        group_pos = [group_pos[0]]
    
    """ Minimum length of the bolt is:
        thickness of end-plate + thickness of column flange + 10 mm to allow
        for two threads over the nut.
        
        To this, the thickness of the nut and washers need to be added.
        This is done in Bolt
    """    
    
    lmin = tp + col.tf + 10 + nut_size[d]["m"] + 2*washer_size[d]["h"]
    #print("Minimum bolt length = {0:4.2f} mm".format(lmin))
    
    bolt = Bolt(d,8.8,lmin)
    
    """ Genate connection """
    conn = EndPlateJoint(col,beam,tp=tp,bp=bp,mat_p=plate_mat,etop=etop,\
                         ebottom=ebottom,bolt=bolt,y_bolts=y_bolts,\
                         e_bolts=e,\
                         bolt_row_pos=bolt_row_pos,
                         groups=bolt_row_groups,
                         group_pos=group_pos,
                         row_types=row_types)
    
    """ Generate welds """
    conn.weld_f = af
    conn.weld_w = aw
    
    conn.workshop = Workshop()
    
    conn.MjEd = MjEd
    conn.VjEd = VjEd
    
    return conn

def suscos_ex(MjEd=0.0,VjEd = 0.0, rows = 4, extended=True):

    steel_grade = 355
    d = 20    
    col = HEA(300,steel_grade)
    beam = IPE(400,fy=steel_grade)
    
    """ Weld size 
        aw .. web
        af .. flange
    """
    aw = 6.0
    af = 9.0
    
    """ Overhang distance
        ebottom .. plate below bottom flange of the beam
        etop .. plate above top flane of the beam
    """
    ebottom = 20.0
    if extended:
        etop = 100.0
    else:
        etop = 20
    
    """ End plate dimensions
        hp .. height
        bp .. width
        tp .. thickness
    """
    plate_mat = "S" + str(steel_grade)
    hp = beam.h + etop + ebottom
    bp = 240.0
    tp = 16.0
    
    
    """ Distance of bolts from the edge of the end plate """    
    e = 60.0   
    
    """ Distances of bolt rows:
        [0] .. distance from top edge of plate
        [i] .. distance between adjacent rows
    """
    if extended:
        p_bolts = [50,110,110,170][:rows]
    else:
        p_bolts = [etop+beam.tf+1.5*af+0.5*d + 20,80,80,120][:rows]


    """ Generate bolt rows
        y is the vertical position of a bolt row. The
        origin is located in the centroid of the beam
    """
    y0 = ebottom + 0.5*beam.h
    y_bolts = []
    y = 0.5*beam.h + etop
    for p in p_bolts:
        y -= p
        y_bolts.append(y)    
    
    """ Generate bolt row position labels """
    if extended:
        bolt_row_pos = [{"flange": END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                        {"flange": INNER_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                        {"flange": INNER_ROW, "plate": OTHER_END_ROW},
                        {"flange": END_ROW, "plate": OTHER_END_ROW},
                        ]
    else:
        bolt_row_pos = [{"flange": END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                        {"flange": INNER_ROW, "plate": OTHER_INNER_ROW},
                        {"flange": INNER_ROW, "plate": OTHER_END_ROW},
                        {"flange": END_ROW, "plate": OTHER_END_ROW},
                        ]
    bolt_row_pos = bolt_row_pos[:rows]
    bolt_row_pos[-1]["flange"] = END_ROW
    
    """ Generate bolt row groups """
    bolt_row_groups = [[0,1], [0,1,2], [1,2]]
    
    if rows == 2:
        bolt_row_groups = [[0,1]]
    
    row_types = [TENSION_ROW,TENSION_ROW,TENSION_ROW,SHEAR_ROW][:rows]
    
    
    # POSITION OF ROWS IN GROUPS
    # For each row, position with respect to column flange and end plate are
    # required.
    if extended:
        group_pos = [[{"flange":END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                      {"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE}
                     ],
                     [{"flange":END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                      {"flange":INNER_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ],
                     [{"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ]
                     ]
    else:
        group_pos = [[{"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ],
                     [{"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                      {"flange":INNER_ROW, "plate": OTHER_INNER_ROW},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ],
                     [{"flange":END_ROW, "plate": OTHER_END_ROW},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ]
                     ]
        
    if rows == 2:
        group_pos = [group_pos[0]]
    
    """ Minimum length of the bolt is:
        thickness of end-plate + thickness of column flange + 10 mm to allow
        for two threads over the nut.
        
        To this, the thickness of the nut and washers need to be added.
        This is done in Bolt
    """    
    
    lmin = tp + col.tf + 10 + nut_size[d]["m"] + 2*washer_size[d]["h"]
    #print("Minimum bolt length = {0:4.2f} mm".format(lmin))
    
    bolt = Bolt(d,10.9,lmin)
    
    """ Genate connection """
    conn = EndPlateJoint(col,beam,tp=tp,bp=bp,mat_p=plate_mat,etop=etop,\
                         ebottom=ebottom,bolt=bolt,y_bolts=y_bolts,\
                         e_bolts=e,\
                         bolt_row_pos=bolt_row_pos,
                         groups=bolt_row_groups,
                         group_pos=group_pos,
                         row_types=row_types)
    
    """ Generate welds """
    conn.weld_f = af
    conn.weld_w = aw
    
    conn.workshop = Workshop()
    
    conn.MjEd = MjEd
    conn.VjEd = VjEd
    
    return conn

def olivaers_ex(MjEd=0.0,VjEd = 0.0, rows = 4, extended=True):

    steel_grade = 275
    d = 16    
    col = HEB(120,steel_grade)
    beam = IPE(300,fy=steel_grade)
    
    """ Weld size 
        aw .. web
        af .. flange
    """
    aw = 4.0
    af = 6.0
    
    """ Overhang distance
        ebottom .. plate below bottom flange of the beam
        etop .. plate above top flane of the beam
    """
    ebottom = 20.0
    if extended:
        etop = 100.0
    else:
        etop = 20
    
    """ End plate dimensions
        hp .. height
        bp .. width
        tp .. thickness
    """
    plate_mat = "S" + str(steel_grade)
    hp = beam.h + etop + ebottom
    bp = col.b
    tp = 8.0
    
    
    """ Distance of bolts from the edge of the end plate """    
    e = 60.0   
    
    """ Distances of bolt rows:
        [0] .. distance from top edge of plate
        [i] .. distance between adjacent rows
    """
    if extended:
        p_bolts = [50,110,110,170][:rows]
    else:
        p_bolts = [etop+beam.tf+1.5*af+0.5*d + 20,80,80,120][:rows]


    """ Generate bolt rows
        y is the vertical position of a bolt row. The
        origin is located in the centroid of the beam
    """
    y0 = ebottom + 0.5*beam.h
    y_bolts = []
    y = 0.5*beam.h + etop
    for p in p_bolts:
        y -= p
        y_bolts.append(y)    
    
    """ Generate bolt row position labels """
    if extended:
        bolt_row_pos = [{"flange": END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                        {"flange": INNER_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                        {"flange": INNER_ROW, "plate": OTHER_END_ROW},
                        {"flange": END_ROW, "plate": OTHER_END_ROW},
                        ]
    else:
        bolt_row_pos = [{"flange": END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                        {"flange": INNER_ROW, "plate": OTHER_INNER_ROW},
                        {"flange": INNER_ROW, "plate": OTHER_END_ROW},
                        {"flange": END_ROW, "plate": OTHER_END_ROW},
                        ]
    bolt_row_pos = bolt_row_pos[:rows]
    bolt_row_pos[-1]["flange"] = END_ROW
    
    """ Generate bolt row groups """
    bolt_row_groups = [[0,1], [0,1,2], [1,2]]
    
    if rows == 2:
        bolt_row_groups = [[0,1]]
    
    row_types = [TENSION_ROW,TENSION_ROW,TENSION_ROW,SHEAR_ROW][:rows]
    
    
    # POSITION OF ROWS IN GROUPS
    # For each row, position with respect to column flange and end plate are
    # required.
    if extended:
        group_pos = [[{"flange":END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                      {"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE}
                     ],
                     [{"flange":END_ROW, "plate": ROW_OUTSIDE_BEAM_TENSION_FLANGE},
                      {"flange":INNER_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ],
                     [{"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ]
                     ]
    else:
        group_pos = [[{"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ],
                     [{"flange":END_ROW, "plate": FIRST_ROW_BELOW_BEAM_TENSION_FLANGE},
                      {"flange":INNER_ROW, "plate": OTHER_INNER_ROW},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ],
                     [{"flange":END_ROW, "plate": OTHER_END_ROW},
                      {"flange":END_ROW, "plate": OTHER_END_ROW}
                     ]
                     ]
        
    if rows == 2:
        group_pos = [group_pos[0]]
    
    """ Minimum length of the bolt is:
        thickness of end-plate + thickness of column flange + 10 mm to allow
        for two threads over the nut.
        
        To this, the thickness of the nut and washers need to be added.
        This is done in Bolt
    """    
    
    lmin = tp + col.tf + 10 + nut_size[d]["m"] + 2*washer_size[d]["h"]
    #print("Minimum bolt length = {0:4.2f} mm".format(lmin))
    
    bolt = Bolt(d,10.9,lmin)
    
    """ Genate connection """
    conn = EndPlateJoint(col,beam,tp=tp,bp=bp,mat_p=plate_mat,etop=etop,\
                         ebottom=ebottom,bolt=bolt,y_bolts=y_bolts,\
                         e_bolts=e,\
                         bolt_row_pos=bolt_row_pos,
                         groups=bolt_row_groups,
                         group_pos=group_pos,
                         row_types=row_types)
    
    """ Generate welds """
    conn.weld_f = af
    conn.weld_w = aw
    
    conn.workshop = Workshop()
    
    conn.MjEd = MjEd
    conn.VjEd = VjEd
    
    return conn


if __name__ == '__main__':
    
    from metku.sections.steel.ISection import HEA, HEB, IPE
    from metku.eurocodes.en1993.en1993_1_8.en1993_1_8 import Bolt
    
    conn = diax_ex_rows(rows=4)
    #conn = example_1()
    #conn = olivaers_ex(extended=True, rows=3)
    
    #conn.bending_resistance(True)
    #conn.rotational_stiffness(True)
    
    #conn.add_stiffener(stiffener='col_top_flange',t=conn.beam.tf)
    #conn.add_stiffener(stiffener='col_bottom_flange',t=conn.beam.tf)
    
    """
    conn.Sj_ini(True)
    conn.rotational_stiffness(True)
    conn.bending_resistance(True)
    """
    
    #ws = Workshop()
    
    #conn.add_stiffener('col_bottom_flange',t=conn.beam.tf)
    #conn.add_stiffener('col_top_flange',t=conn.beam.tf)
    
    #conn.info(draw=True)
    
    #fig_dir = 'P:\\INTRA_HOME\\Tutkimus\\IXConnections\\figures\\'
    #conn.draw(fig_dir + "testfig.svg")
    
    #conn.bending_resistance(True)
    #conn.rotational_stiffness(True)
    #conn.shear_resistance(True)
    #conn.cost(verb=True)
    """
    for row in conn.bolt_rows:
        print(row.stiffness_factors)
    """
    """
        for group in row.groups:
            group.Tstub_col_flange.leff_row(row)
            group.Tstub_end_plate.leff_row(row)
    """
    """
    for group in conn.row_groups:
        print(group.Tstub_col_flange._leff_rows)
        print(group.Tstub_end_plate._leff_rows)
    """
    
    #conn.row_groups[0].Tstub_col_flange.leff_nc_row(r)
    #conn.row_groups[0].Tstub_col_flange.leff_cp_row(r)
        
        


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 30 08:16:41 2020

@author: kmela
"""
import numpy as np

class CrossSection():
    """ Base class for generic cross-sections
    
    Specific cross sections are subclasses of this class
    
    NOTE: The original purpose of this class is to act as a generic
            cross-section to be used in structural optimization
    
    """

    # Constructor
    def __init__(self,     
                 material=None,
                 A=0,
                 I=[0,0]):
        
        self.material = material
        self.A = A
        self.I = np.asarray(I)       

    def __repr__(self):
        return type(self).__name__
    
    @property
    def E(self):
        return self.material.E


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Material data

Created on Wed Aug  7 09:23:08 2019

@author: kmela
"""

from math import sqrt
import metku.eurocodes.en1993.en1993_1_2 as en1993_1_2

steels = {"S235": {"fy": 235.0, "fu": 360.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "S275": {"fy": 275.0, "fu": 430.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "S355": {"fy": 355.0, "fu": 510.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "S420": {"fy": 420.0, "fu": 500.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "S460M": {"fy": 460.0, "fu": 540.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "S460Q": {"fy": 460.0, "fu": 570.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "B500B": {"fy": 500.0, "fu": 550.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "S220GD": {"fy": 220.0, "fu": 300.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "S250GD": {"fy": 250.0, "fu": 330.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "S280GD": {"fy": 280.0, "fu": 360.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "S320GD": {"fy": 320.0, "fu": 390.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "S350GD": {"fy": 350.0, "fu": 420.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
        "S355MC": {"fy": 355.0, "fu": 430.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "S500": {"fy": 500.0, "fu": 550.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "S700": {"fy": 700.0, "fu": 750.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "DP1000": {"fy": 590.0, "fu": 980.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       "HCT980X": {"fy": 590.0, "fu": 980.0, "E": 210000.0, "v": 0.3, "rho": 7850e-9},
       }

stainless = {'1.4301': {'cold_strip': {'fy': 230, 'fu': 540}, 
                        'hot_strip': {'fy': 210, 'fu': 520},
                        'hot_plate': {'fy': 210, 'fu': 520},
                        'sections': {'fy': 190, 'fu': 500},
                        'E': 200000.0, 'v':0.3, 'rho': 7850e-9}}

class Steel:
    
    def __init__(self,steel_grade="S355"):
        """ Constructor for Steel material class """

        self.name = steel_grade
        self.fy = steels[steel_grade]["fy"]
        self.fu = steels[steel_grade]["fu"]
        self.E = steels[steel_grade]["E"]
        self.nu = steels[steel_grade]["v"]
        self.rho = steels[steel_grade]["rho"]
        self.G = self.E/2/(1+self.nu)
    
    def __repr__(self):
        
        return f"{self.name}"
    
    @property
    def young(self):
        """ Young's modulus: alternative way of calling 'E' """
        return self.E
    
    def eps(self):
        return sqrt(235.0/self.fy)
    
    def Etemp(self,t):
        """ Modification of Young's modulus according to
            EN 1993-1-2
        """
        return en1993_1_2.EaT(t,self.E)
    
    def fy_temp(self,t):
        """ Modification of yield strength according to
            EN 1993-1-2
        """
        return en1993_1_2.fyT(t,self.fy)
    
    def fp_temp(self,t):
        """ Modification of proportional limit according to
            EN 1993-1-2
        """
        return en1993_1_2.fpT(t,self.fy)
    
    def fu_temp(self,t):
        """ Modification of ultimate strength according to
            EN 1993-1-2
        """
        return en1993_1_2.fuT(t,self.fy_temp(t))
        
    
    def stress_temp(self,strain,t,hardening=False):
        """ Calculate stress in steel for given strain at temperature
            't', possiblity with hardenin effect included.
        """
        return en1993_1_2.stress(strain,t,self.E,self.fy,hardening)
    
class StainlessSteel:
    
    def __init__(self,steel_grade="1.4301",product='sections'):
        """ Constructor """ 
        self.name = steel_grade
        self.product = product
        
        self.fy = stainless[steel_grade][product]["fy"]
        self.fu = stainless[steel_grade][product]["fu"]
        self.E = stainless[steel_grade]["E"]
        self.nu = stainless[steel_grade]["v"]
        self.rho = stainless[steel_grade]["rho"]
        self.G = self.E/2/(1+self.nu)
    
    def __repr__(self):
        
        return f"{self.name}"
    
    def eps(self):
        return sqrt(235.0/self.fy*self.E/200000.0)


s = Steel()
s.fy